
#-----------------------------------------------------------------------
# Override some Tk widget defaults and define some configuration
# parameters that are used elsewhere is this script.
#-----------------------------------------------------------------------
proc set_defaults {} {

# Change the default colors of certain Tcl/Tk widgets.

  option add *background \#E5E5E5
  option add *foreground black
  option add *Entry*Background \#B0E2FF
  option add *Entry*Foreground black
  option add *Text*Background white
  option add *Text*Foreground pink
  option add *Text*Foreground black
  option add *selectColor yellow

# Reduce the padding around the text of buttons.

  option add *Button.padX 1m
  option add *Button.padY 0.5m
  option add *Checkbutton.padY 0
  option add *Checkbutton.BorderWidth 1
  option add *Entry.BorderWidth 1
  option add *Text.BorderWidth 0
  option add *Label.BorderWidth 1

# Override the default font with a smaller one, so that the window will
# fit!

  option add *Font {Helvetica -10 normal} widgetDefault

# Set the colors color that are used for the frame-title bars.

  set ::bar_bg \#999999
  set ::bar_fg turquoise

# Specify the number of calibration-diode state machine steps to
# display.

  set ::max_ncal 10
}

#-----------------------------------------------------------------------
# Create a new frame containing phase-switch configuration fields.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc scan_cnf_phase_frame {w} {

# Create the frame that encloses the configuration area.

  frame $w

# Set the width of each of the configuration entry widgets.

  set width 14

# The phase-switch configuration area has a title at the left, followed by
# fields for each of the configuration parameters.

  label $w.t -text {Phase Switch configuration} -fg $::bar_fg -bg $::bar_bg -relief raised
  titled_a_b_checkbutton_frame $w.as "Active\nswitches" left 1 1
  titled_a_b_checkbutton_frame $w.is "Initially-closed\nswitches" left 0 0
  titled_entry_frame $w.sps "Samples per\nphase-switch state" $width 250

# Pack the fields from left to right, with the title placed at the
# same vertical position as the entry field.

  pack $w.t -side top -anchor nw -anchor center -fill x
  pack $w.as $w.is $w.sps -side left -fill x -anchor s -expand true

  return $w
}

#-----------------------------------------------------------------------
# Create a new frame containing cal-diode configuration fields.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc scan_cnf_cal_frame {w} {

# Create the frame that encloses the configuration area.

  frame $w

# Set the width of each of the configuration entry widgets.

  set width 16

# The cal-diode configuration area has a title at the left, followed by
# fields for each of the configuration parameters.

  label $w.t -text {Cal diode configuration} -fg $::bar_fg -bg $::bar_bg -relief raised

# Create a frame in which the number of calibration steps can be entered.

  titled_entry_frame $w.ncal "Number of\ncalibration\nsteps" 5 1

# Create a frame containing row labels for the table of calibration
# steps.

  frame $w.r
  label $w.r.a -text {Diode A state}
  label $w.r.b -text {Diode B state}
  label $w.r.n -text {Integrations}
  pack $w.r.a $w.r.b $w.r.n

# Create a frame containing the calibration-diode state-machine steps.

  cal_diode_state_machine_frame $w.sm $::max_ncal

# Pack the fields from left to right, with the title placed at the
# same vertical position as the entry field.

  pack $w.t -side top -anchor center -fill x
  pack $w.ncal $w.r $w.sm -side left -fill x -anchor s -expand true

  return $w
}

#-----------------------------------------------------------------------
# Create a new frame containing a list of a specified number of cal-diode
# state fields. Each field contains two checkbuttons, specifying whether
# diodes A and B are switched on, plus an entry field specifying the
# duration of the calibration state as a number of integrations. The
# name of the widget of the i'th state (i=0..$n-1) is $w.$i.
#
# Arguments:
#    w         The pathname to give the new frame.
#    n         The number of calibration steps to create.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc cal_diode_state_machine_frame {w n} {

# Create the frame that encloses the table of steps.

  frame $w

# Create the array of steps.

  cal_diode_state_frame $w.0 false false 1
  for {set i 1} {$i <= $n} {incr i} {
    cal_diode_state_frame $w.$i false false 1
    pack $w.$i -side left
  }

  return $w
}

#-----------------------------------------------------------------------
# Create a new frame containing a single cal-diode state-machine step.
# This contains two checkbuttons, specifying whether diodes A and B
# are switched on, plus an entry field specifying the duration of the
# calibration state as a number of integrations.
#
# The variables that represent the checkbuttons are named ::$w.a and
# ::$w.b respectively, and the variable that represents the value of
# the entry widget is named ::$w.e.
#
# Arguments:
#    w         The pathname to give the new frame.
#    a         The initial state of button A (true or false).
#    b         The initial state of button B (true or false).
#    n         The initial duration to place in the entry field.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc cal_diode_state_frame {w a b n} {

# Create the frame that encloses the configuration area.

  frame $w -relief ridge -bd 2

# Create the frame within which the checkbuttons will be arranged.

  frame $w.cb

# Create the two checkbuttons.

  checkbutton $w.a -text {A} -variable $w.a -pady 0
  set ::$w.a $a
  checkbutton $w.b -text {B} -variable $w.b -pady 0
  set ::$w.b $b

# Stack the buttons vertically.

  pack $w.a $w.b -side top -anchor center -pady 0

# Now create an entry field for the number of integrations.

  entry $w.e -width 3
  replace_entry_value $w.e $n

# Pack the diode-selection and duration fields one above the other.

  pack $w.cb $w.e -side top -pady 0

  return $w
}

#-----------------------------------------------------------------------
# Create a new frame containing timing configuration fields.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc scan_cnf_timing_frame {w} {

# Create the frame that encloses the configuration area.

  frame $w

# Set the width of each of the configuration entry widgets.

  set width 8

# The cal-diode configuration area has a title at the left, followed by
# fields for each of the configuration parameters.

  label $w.t -text {Timing configuration} -bg $::bar_bg -relief raised -fg $::bar_fg
  titled_entry_frame $w.ps "Phase switch\ndt (100ns)" $width 10
  titled_entry_frame $w.dr "Diode rise\ndt (100ns)" $width 1000
  titled_entry_frame $w.df "Diode fall\ndt (100ns)" $width 1000
  titled_entry_frame $w.ip "Integ period\n(cycles)" $width 10000
  titled_entry_frame $w.rt "Round-trip\ndt (100ns)" $width 10
  titled_entry_frame $w.ho "Holdoff\ndt (25.6us)" $width 7

# Pack the fields from left to right, with the title placed at the
# same vertical position as the entry field.

  pack $w.t -side top -anchor center -fill x
  pack $w.ps $w.dr $w.df $w.ip $w.rt $w.ho -side left -fill x -anchor s -expand true

  return $w
}

#-----------------------------------------------------------------------
# Create a new frame containing sampler configuration fields.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc scan_cnf_sampler_frame {w} {

# Create the frame that encloses the configuration area.

  frame $w

# Set the width of each of the configuration entry widgets.

  set width 8

# The has a heading above the fields of each of the configuration parameters.

  label $w.t -text {Sampler configuration} -bg $::bar_bg -relief raised -fg $::bar_fg

# Encapsulate the test-samples option-menu and label in a frame.

  option_menu $w.type {Type of input samples: } [list {ADC} {Fake}] {}
  set ::$w {ADC}

# Pack the fields from left to right, with the title placed at the
# same vertical position as the entry field.

  pack $w.t -side top -anchor center -fill x
  pack $w.type -side left -fill x -anchor s -expand true

  return $w
}

#-----------------------------------------------------------------------
# Create a labelled option menu.
#
# The value of the option menu can be changed or read via a local alias
# to the global $w variable that is created by this command.
# The label will be named $w.l and the menubutton $w.opt.
#
# Input:
#  w       The path name to give the frame that surrounds the label
#          and menu button.
#  label   A label to be placed to the left of the option menu button.
#  values  A list of option names. The initial value of the option menu
#          will be the first element of this list.
#  cmd     A command to be executed whenever the selected option menu
#          entry changes. The newly selected value will be inserted as
#          a single list element wherever %v occurs in $cmd. If no
#          command is required, pass {} here.
#  args    Extra configuration arguments for the menu button.
# Output:
#  return  The name of the widget ($w).
#-----------------------------------------------------------------------
proc option_menu {w label values cmd args} {
#
# Encapsulate the label and menubutton within a frame.
#
  frame $w
  label $w.l -text $label
#
# Create an option menubutton and the menu that it will post.
#
  eval {menubutton $w.opt -menu $w.opt.m -indicator on -relief raised -bd 2} $args
  set m [menu $w.opt.m -tearoff no]
#
# Insert the list of options as radio buttons linked to a global variable
# that has the same name as the menubutton widget.
#
  foreach value $values {
    $m add radiobutton -label $value -value $value -variable ::$w
  }
#
# Set the initial value of the menu and install this as the initial
# menubutton label.
#
  set ::$w [lindex $values 0]
  option_menu_handler {} ::$w
  pack $w.l $w.opt -side left
#
# Link the command that will both keep the menubutton label up to date
# and evaluate the user's callback, to updates of the radiobutton variable.
#
  trace variable ::$w w [list option_menu_handler $cmd ::$w]
  return $w
}

#-----------------------------------------------------------------------
# This is a private callback of the option_menu procedure. Whenever the
# selected option menu value changes, it updates the label of the
# associated option menu button to match and invokes the user's callback.
#
# Input:
#  cmd       The user's callback command.
#  w         The path name of the frame that surrounds the menubutton
#            widget. This is also the name of the global
#            variable that contains the value of the option menu.
#  args      Ignored extra arguments. This allows the procedure to
#            be used as a trace callback.
#-----------------------------------------------------------------------
proc option_menu_handler {cmd w args} {
  upvar #0 $w var
  $w.opt configure -text $var
  if {[string compare $cmd {}] != 0} {
    regsub -all {%v} $cmd [list $var] newcmd
    eval $newcmd
  }
}

#-----------------------------------------------------------------------
# Create a new frame containing scan command-initiation buttons.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc scan_cmd_frame {w} {

# Create the frame that encloses the configuration area.

  frame $w

  label $w.t -text {Scan initiation commands} -bg $::bar_bg -relief raised -fg $::bar_fg
  pack $w.t -side top -anchor center -fill x

  # Create all fields associated with the start-scan command.

  set start [frame $w.start]

  button $start.send -text "Start scan" -command "send_start_scan \[$start.mjd.when get\] \[$start.time.when get\]"
  label $start.l1 -text "at MJD"

  # Create a pair of widgets, one to display the current MJD and a second
  # in which the user can enter the target MJD.

  set mjd [frame $start.mjd]
  label $mjd.now
  entry $mjd.when -width 6
  replace_entry_value $mjd.when [lindex [ccb get time] 0]
  pack $mjd.now $mjd.when -side left -anchor w

  label $start.l2 -text "time"

  # Create a pair of widgets, one to display the current time and a second
  # in which the user can enter the target time.

  set time [frame $start.time]
  label $time.now
  entry $time.when -width 5
  replace_entry_value $time.when [lindex [ccb get time] 1]
  pack $time.now $time.when -side left -anchor w

  label $start.l3 -text "seconds"

  # Justify the start button and its parameter fields to the left
  # of the frame.

  pack $start.send $start.l1 $start.mjd $start.l2 $start.time $start.l3 -side left -anchor s

  # Create the fields of the stop-scan command.

  button $w.stop -text "Stop scan" -command send_stop_scan

  # Pack the start-scan and stop-scan areas at opposite ends of the
  # frame.

  pack $w.start -side left -anchor s
  pack $w.stop -side right -anchor s

  return $w
}

#-----------------------------------------------------------------------
# Create a new frame containing dump-mode command-initiation buttons.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc dump_cmd_frame {w} {

# Create the frame that encloses the configuration area.

  frame $w

  label $w.t -text {Dump-mode control commands} -bg $::bar_bg -relief raised -fg $::bar_fg
  pack $w.t -side top -anchor center -fill x

  # Create a frame to enclose the dump-command button and its parameter
  # entry widgets.
 
  set dump [frame $w.dump -bd 1 -relief ridge]

  # Create the button which initiates a dump-scan.

  button $dump.send -text "Dump" -command "send_dump_scan \[$dump.adc.e get\] \[$dump.samples.e get\] \[$dump.frames.e get\]"

  # Provide entry areas for the parameters of the dump-scan command.

  labeled_entry_frame $dump.adc {ADC} 6 0
  labeled_entry_frame $dump.samples {samples/frame} 6 max
  labeled_entry_frame $dump.frames {max frames} 6 all

  pack $dump.send $dump.adc $dump.samples $dump.frames -side left

  pack $dump -side left

  return $w
}

#-----------------------------------------------------------------------
# Create a new frame containing server connection buttons.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc link_cmd_frame {w} {

# Create the frame that encloses the configuration area.

  frame $w

  label $w.t -text {Link control} -bg $::bar_bg -relief raised -fg $::bar_fg
  pack $w.t -side top -anchor center -fill x

  # Create the fields associated with the "connect" command.

  button $w.connect -text "Connect" -command "connect_to_server \[$w.host get\]"
  label $w.l1 -text "to:"
  entry $w.host -textvariable $w.host -width 16
  bind $w.host <Return> "$w.connect invoke"
  label $w.l2 -text "  "

  # Justify the connect button and its parameter fields to the left
  # of the frame.

  pack $w.connect $w.l1 $w.host $w.l2 -side left

  # Create the buttons that send disconnect, ping and status-request commands.

  button $w.disconnect -text "Disconnect" -command "ccb disconnect"
  button $w.ping -text "Send ping" -command send_ping_command
  button $w.status -text "Request status" -command "ccb send status_request"

  # Pack the buttons such that they fill up the horizontal width of the frame.

  pack $w.disconnect $w.ping $w.status -side left -fill x -expand true

  return $w
}

#-----------------------------------------------------------------------
# Create a new frame containing system control buttons.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc system_cmd_frame {w} {

# Create the frame that encloses the configuration area.

  frame $w

  label $w.t -text {System control} -bg $::bar_bg -relief raised -fg $::bar_fg
  pack $w.t -side top -anchor center -fill x

  # Create buttons for the reset, reboot and shutdown commands.

  button $w.reset -text "Reset" -command send_reset_command
  button $w.reboot -text "Reboot" -command "ccb send reboot"
  button $w.shutdown -text "Shutdown" -command "ccb send shutdown"

# Encapsulate the test-samples option-menu and label in a frame.

  option_menu $w.load {Load driver: } [list {Normal} {Virtual}] reconfigure_ccb
  set ::$w.load {Virtual}

  # Pack the buttons such that they fill up the horizontal width of the frame.

  pack $w.reset $w.reboot $w.shutdown $w.load -side left -fill x -expand true

  return $w
}

#-----------------------------------------------------------------------
# Switch the CCB into the off state.
#-----------------------------------------------------------------------
proc send_off_streams {} {
  ccb send telemetry "log_stream"
  .tel.off configure -bg red
  .tel.standby configure -bg [.tel cget -bg]
  .tel.ready configure -bg [.tel cget -bg]
}

#-----------------------------------------------------------------------
# Switch the CCB into the standby state.
#-----------------------------------------------------------------------
proc send_standby_streams {} {
  ccb send telemetry "log_stream monitor_stream"
  .tel.off configure -bg [.tel cget -bg]
  .tel.standby configure -bg yellow
  .tel.ready configure -bg [.tel cget -bg]
}

#-----------------------------------------------------------------------
# Switch the CCB into the ready state.
#-----------------------------------------------------------------------
proc send_ready_streams {} {
  ccb send telemetry "all_streams"
  .tel.off configure -bg [.tel cget -bg]
  .tel.standby configure -bg [.tel cget -bg]
  .tel.ready configure -bg green
}

#-----------------------------------------------------------------------
# Send a reset command to the CCB.
#-----------------------------------------------------------------------
proc send_reset_command {} {
  ccb send reset

# Replace the default CCB configuration with the parameters that are
# currently shown by the GUI.

  reconfigure_ccb
}

#-----------------------------------------------------------------------
# Send a ping command to the CCB.
#-----------------------------------------------------------------------
proc send_ping_command {} {
  ccb send ping

# The following statement is only executed when above command
# doesn't complain that the preceding ping hasn't been received.

  report notice {A response was received to the previous ping command}
}

#-----------------------------------------------------------------------
# Create a new frame containing telemetry-control buttons.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc telemetry_cmd_frame {w} {

# Create the frame that encloses the configuration area.

  frame $w

  label $w.t -text {Telemetry control} -bg $::bar_bg -relief raised -fg $::bar_fg
  pack $w.t -side top -anchor center -fill x

  # Create the buttons used to switch the CCB into "off", "standby" and
  # "ready" states.

  button $w.off -text "Off" -command send_off_streams
  button $w.standby -text "Standby" -command send_standby_streams
  button $w.ready -text "Ready" -command send_ready_streams

  # Create the fields associated with the monitoring control command.

  button $w.monitor -text "Configure\nmonitoring" -command send_monitor_config
  label $w.l3 -text "period:"
  entry $w.monitor_dt -width 5
  replace_entry_value $w.monitor_dt 10
  label $w.l4 -text "integ's  "

  # Create the fields of the log control command.

  button $w.log -text "Configure\nlogging" -command "ccb send logger \[$w.log_dt get\]"
  label $w.l5 -text "purge\nperiod:"
  entry $w.log_dt -width 3
  replace_entry_value $w.log_dt 2
  label $w.l6 -text "s"

  # Display the above fields left-to-right.

  pack $w.off $w.standby $w.ready $w.monitor $w.l3 $w.monitor_dt $w.l4 $w.log $w.l5 $w.log_dt $w.l6 -side left

  return $w
}

#-----------------------------------------------------------------------
# Send a monitor configuration command to the CCB server.
#-----------------------------------------------------------------------
proc send_monitor_config {} {
  ccb send monitor [.tel.monitor_dt get]
}

#-----------------------------------------------------------------------
# Send a log configuration command to the CCB server.
#-----------------------------------------------------------------------
proc send_log_config {} {
  ccb send logger [.tel.log_dt get]
}

#-----------------------------------------------------------------------
# Create the frame that is used to display log messages.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc log_display_frame {w} {

# Create the frame that encloses the area.

  frame $w

  label $w.t -text {Log messages} -bg $::bar_bg -relief raised -fg $::bar_fg
  pack $w.t -side top -anchor center -fill x

  # Create the frame in which the text widget and its scroll bar will be
  # aggregated.

  set msg [frame $w.msg]
  
  # Create the text widget in which log messages will be displayed.

  text $msg.text -bg beige -height 5 -yscrollcommand "$msg.scroll set" -takefocus 0 -state disabled

  # Create display styles for each of the log levels.

  foreach t {{info black} {notice black} {warning orange} {error red} {fault red} {fatal red}} {
    $msg.text tag configure [lindex $t 0] -foreground [lindex $t 1]
  }

  # Create a vertical scroll bar configured to scroll the text widget.

  scrollbar $msg.scroll -orient vertical -command "$msg.text yview"

  # Display the text widget and its scrollbar side by side, with the text
  # widget expanding horizontally to take up all space not taken by the
  # scrollbar.

  pack $msg.text -expand true -fill x -side left
  pack $msg.scroll -side left -fill y

  # Display the log area below its title.

  pack $msg -side top -fill x

  return $w
}

#-----------------------------------------------------------------------
# Create the frame that is used to display integrated data.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc integ_display_frame {w} {

# Create the frame that encloses the area.

  frame $w 

  label $w.t -text {The latest integrated data} -bg $::bar_bg -relief raised -fg $::bar_fg
  pack $w.t -side top -anchor center -fill x

  # Create a message area to display the formatted integrated data,
  # requesting a width sufficient to display 8 32-bit space-separated
  # integers.

  message $w.msg -width 100c -bg beige -font {Helvetica -9 normal}
  pack $w.msg -side top -fill x

  return $w
}

#-----------------------------------------------------------------------
# Create the frame that is used to display monitor data.
#
# Arguments:
#    w         The pathname to give the new frame.
# Return value:
#    $w
#-----------------------------------------------------------------------
proc monitor_display_frame {w} {

# Create the frame that encloses the area.

  frame $w

  label $w.t -text {The latest monitor data} -bg $::bar_bg -relief raised -fg $::bar_fg
  pack $w.t -side top -anchor center -fill x

  # Create a message area to display the formatted integrated data,
  # requesting a width sufficient to display 8 32-bit space-separated
  # integers.

  message $w.msg -width 100c -bg beige -font {Helvetica -8 normal}
  pack $w.msg -side top -fill x

  return $w
}

#-----------------------------------------------------------------------
# Create a frame containing a title stacked above an entry widget.
#
# Arguments:
#  w           The pathname to give the frame.
#  title       The title to give the field.
#  width       The width of the entry field.
#  value       The initial value displayed in the entry widget.
# Return value:
#  $w
#-----------------------------------------------------------------------
proc titled_entry_frame {w title width value} {

# Create the frame which encloses the title and the entry field.

  frame $w

# Create the title.

  label $w.t -text $title

# Create the entry widget.

  entry $w.e -width $width
  replace_entry_value $w.e $value

# Pack the title and entry widget vertically.

  pack $w.t $w.e -side top -anchor center

  return $w
}

#-----------------------------------------------------------------------
# Create a frame containing a label to the left of an entry widget.
#
# Arguments:
#  w           The pathname to give the frame.
#  text        The label to give the field.
#  width       The width of the entry field.
#  value       The initial value displayed in the entry widget.
# Return value:
#  $w
#-----------------------------------------------------------------------
proc labeled_entry_frame {w text width value} {

# Create the frame which encloses the title and the entry field.

  frame $w

# Create the title.

  label $w.t -text $text

# Create the entry widget.

  entry $w.e -width $width
  replace_entry_value $w.e $value

# Pack the label and entry widget from left to right.

  pack $w.t $w.e -side left -anchor center

  return $w
}

#-----------------------------------------------------------------------
# Create a frame containing a title stacked above two vertically
# stacked checkbutton widgets, labelled a and b. The variables that
# record the values of these buttons are ::$w.a and ::$w.b
# respectively.
#
# Arguments:
#  w           The pathname to give the frame.
#  title       The title to give the field.
#  side        The side of the checkbutton frame to pack the
#              checkbuttons against (left or top).
#  a           The value to give checkbutton a (0 or 1).
#  b           The value to give checkbutton b (0 or 1).
# Return value:
#  $w
#-----------------------------------------------------------------------
proc titled_a_b_checkbutton_frame {w title side a b} {

# Create the frame which encloses the title and the check-button areas.

  frame $w

# Create the title.

  label $w.t -text $title

# Create the frame within which the checkbuttons will be laid out.

  frame $w.cb

# Create the two checkbuttons.

  checkbutton $w.cb.a -text {A} -variable $w.a
  checkbutton $w.cb.b -text {B} -variable $w.b

# Establish the initial states of the checkbuttons.

  set_a_b_checkbuttons $w $a $b

# Stack the checkbuttons horizontally.

  pack $w.cb.a $w.cb.b -side $side

# Stack the title and checkbutton area vertically.

  pack $w.t $w.cb -side top -anchor center

  return $w
}

#-----------------------------------------------------------------------
# Get the current states of a pair of checkbuttons created by
# titled_a_b_checkbutton_frame.
#
# Input:
#  w      The pathname of a widget created by titled_a_b_checkbutton_frame
# Input/Output:
#  a b    These should be the names of two variables in the calling
#         procedure. These variables will be assigned the current values
#         of the A and B checkbuttons.
#-----------------------------------------------------------------------
proc read_a_b_checkbuttons {w a b} {
  uplevel {
    set a [set ::$w.a]
    set b [set ::$w.b]
  }
}

#-----------------------------------------------------------------------
# According to the currently selected members of a pair of a and b
# check-buttons created by the titled_a_b_checkbutton_frame procedure
# generate the corresponding textual list of selected phase switches.
#
# Input:
#  w      The pathname of a widget created by titled_a_b_checkbutton_frame
# Output:
#  return The set of phase switches in the form required by the CCB Tcl
#         interface.
#-----------------------------------------------------------------------
proc read_switch_buttons {w} {

# Get the states of the checkbuttons that represent switches a and b.

  read_a_b_checkbuttons $w a b

# Switch A selected?

  if {$a} {

# Switch B also selected?

    if {$b} {
      return "all_switches"
    } else {
      return "switch_a"
    }

# Just switch B?

  } elseif {$b} {
    return "switch_b"

# No switches selected?

  } else {
    return "no_switches"
  }
  return "oh dear"
}

#-----------------------------------------------------------------------
# According to the currently selected members of a pair of a and b
# check-buttons created by the titled_a_b_checkbutton_frame procedure
# generate the corresponding textual list of selected cal-diodes.
#
# Input:
#  w      The pathname of a widget created by titled_a_b_checkbutton_frame
# Output:
#  return The set of cal-diodes in the form required by the CCB Tcl
#         interface.
#-----------------------------------------------------------------------
proc read_cal_diodes {w} {

# Get the states of the checkbuttons that represent cal-diodes a and b.

  read_a_b_checkbuttons $w a b

# Switch A selected?

  if {$a} {

# Diode B also selected?

    if {$b} {
      return "all_diodes"
    } else {
      return "diode_a"
    }

# Just diode B?

  } elseif {$b} {
    return "diode_b"

# No diodes selected?

  } else {
    return "no_diodes"
  }
}

#-----------------------------------------------------------------------
# Set which of a pair of a and b check-buttons created by the
# titled_a_b_checkbutton_frame procedure are enabled.
#
# Input:
#  w      The pathname of a widget created by titled_a_b_checkbutton_frame
#  a      The value (0 or 1) to give button a.
#  b      The value (0 or 1) to give button b.
#-----------------------------------------------------------------------
proc set_a_b_checkbuttons {w a b} {
  set ::$w.a $a
  set ::$w.b $b
}

#-----------------------------------------------------------------------
# Update the scan configuration, then send a start-scan command to the CCB.
#
# Arguments:
#  mjd         The Modified Julian Day number at which to start the scan.
#  time        The time of day at which to start the scan (seconds into day).
#-----------------------------------------------------------------------
proc send_start_scan {mjd time} {

  # Send the configuration of the new scan.

  update_scan_config
  
  # Send the start-scan command.

  ccb send start_scan $mjd $time
}

#-----------------------------------------------------------------------
# Update the scan configuration, then send a stop-scan command to the CCB.
#-----------------------------------------------------------------------
proc send_stop_scan {} {

  # Send the configuration of the new scan.

  update_scan_config
  
  # Send the stop-scan command.

  ccb send stop_scan
}

#-----------------------------------------------------------------------
# Update the scan configuration, then send a dump-scan command to the CCB.
#
# Arguments:
#  adc       The number of the ADC to dump.
#  samples   The maximum number of samples/frame to dump.
#  frames    The number of integration frames to dump.
#-----------------------------------------------------------------------
proc send_dump_scan {adc samples frames} {

  # Send the configuration of the new scan.

  update_scan_config
  
  # Send the dump-scan command.

  ccb send dump_scan $adc $samples $frames
}

#-----------------------------------------------------------------------
# Send the CCB Tcl/interface an update of the current scan configuration.
#-----------------------------------------------------------------------
proc update_scan_config {} {

  # Update the phase switch configuration.

  ccb configure phase_switches [read_switch_buttons .phase.as] [read_switch_buttons .phase.is] [.phase.sps.e get]

  # Update the timing configuration.

  set w .timing
  ccb configure timing [$w.ps.e get] [$w.dr.e get] [$w.df.e get] [$w.ip.e get] [$w.rt.e get] [$w.ho.e get]

  # Update the sampler configuration.

  switch [set ::.sampler.type] {
    ADC {set type adc_samples}
    Fake {set type fake_samples}
    default {set type adc_samples}
  }
  ccb configure sampler $type

  # Get the number of steps in the calibration-diode state machine.

  set w .cal
  set ncal [$w.ncal.e get]
  if { ! [is_uint $ncal] || $ncal > $::max_ncal } {
    error "Invalid number of calibration-diode steps"
  }

  # Form the lists of cal-diode state parameters.

  set states {}
  set times {}
  for {set i 1} {$i <= $ncal} {incr i} {
    set s $w.sm.$i
    lappend states [read_cal_diodes $s]
    lappend times [$s.e get]
  }

  # Update the cal-diode configuration.

  ccb configure cal_diode $ncal $states $times
}

#-----------------------------------------------------------------------
# Establish a new connection to a CCB server, and send it the current
# configuration parameters.
#
# Arguments:
#   host       The address of the server.
#-----------------------------------------------------------------------
proc connect_to_server {host} {

  ccb connect $host

# Replace the default CCB configuration with the parameters that are
# currently shown by the GUI.

  reconfigure_ccb
}

#-----------------------------------------------------------------------
# This procedure is called whenever a new connection is established or
# a reset command is sent, in order to install the configuration
# currently selected by the user.
#-----------------------------------------------------------------------
proc reconfigure_ccb {} {

# Install the current configuration.

  send_off_streams
  send_monitor_config
  send_log_config

# Load the currently selected driver.

  switch [set ::.sys.load] {
    Normal {set driver normal}
    Virtual {set driver virtual}
    default {set driver normal}
  }
  ccb send load_driver $driver

# Start an intra-scan with the current parameters.

  send_stop_scan
}

#-----------------------------------------------------------------------
# Append a message to the log message window.
#
# Arguments:
#  type     The type of error message, which must be one of:
#              info  notice  warning  error  fault  fatal
#  msg      The message to be displayed.
#-----------------------------------------------------------------------
proc report {type msg} {
  set w .log.msg.text
  $w configure -state normal
  $w insert end "$type: $msg\n" $type
  $w see end
  $w configure -state disabled
}

#-----------------------------------------------------------------------
# This is a log-message callback script.
#-----------------------------------------------------------------------
proc ccb_log_callback {} {
  set data [ccb get log]
  report [lindex $data 4] "\[[lindex $data 3]\] [lindex $data 5]"
}

#-----------------------------------------------------------------------
# This is a CCB-status callback script.
#-----------------------------------------------------------------------
proc ccb_status_callback {} {
  report info "CCB status report: [ccb get status]"
}
 
#-----------------------------------------------------------------------
# This is a CCB command-error callback script.
#-----------------------------------------------------------------------
proc ccb_error_callback {} {
  set data [ccb get error]
  report error "CCB command error: $data"
}

#-----------------------------------------------------------------------
# This is a CCB integrated-data callback script.
#-----------------------------------------------------------------------
proc ccb_integ_callback {} {
  set data [ccb get integ]
  set msg "mjd=[lindex $data 0] sec=[format "%.3f" [expr [lindex $data 1] + [lindex $data 2] / 1e9]] scan=[lindex $data 3] id=[lindex $data 4] flags=\[[lindex $data 5]\]\n"
  set i 0
  foreach n [lrange $data 7 end] {
    append msg [format { %10.10u} $n]
    if {[incr i] % 10 == 0} {append msg "\n"}
  }
  .integ.msg configure -text $msg
}

#-----------------------------------------------------------------------
# This is a CCB monitor-data callback script.
#-----------------------------------------------------------------------
proc ccb_monitor_callback {} {
  set data [ccb get monitor]
  set msg "mjd=[lindex $data 0] sec=[format "%.3f" [expr [lindex $data 1] + [lindex $data 2] / 1e9]] scan=[lindex $data 3] id=[lindex $data 4]\n"
  set i 0
  foreach n [lrange $data 6 end] {
    append msg [format { %10.10u} $n]
    if {[incr i] % 10 == 0} {append msg "\n"}
  }
  .monitor.msg configure -text $msg
}

#-----------------------------------------------------------------------
# This procedure is called every time to update the displayed date and
# time.
#-----------------------------------------------------------------------
proc redisplay_time {} {
  set time [ccb get time]
  .scan.start.mjd.now configure -text "([lindex $time 0]):"
  .scan.start.time.now configure -text "([lindex $time 1]):"
  after 1000 redisplay_time
}

#-----------------------------------------------------------------------
# Arrange that errors that aren't explicitly caught by this script, are
# displayed in the log window.
#
# Arguments:
#  message
#-----------------------------------------------------------------------
proc bgerror {message} {
  if {$message != ""} {
    report error $message
  }
}

#-----------------------------------------------------------------------
# Return 1 if the argument is an unsigned integer.
#-----------------------------------------------------------------------
proc is_uint {arg} {
  regexp "^\[ \t]*\[0-9]+\[ \t]*\$" $arg
}

#-----------------------------------------------------------------------
# Replace the current value of an entry widget with a new value.
#
# Input:
#  w      The pathname of the entry widget.
#  v      The new value.
#-----------------------------------------------------------------------
proc replace_entry_value {w v} {
  $w delete 0 end
  $w insert 0 $v
}

# Set the default fonts and colors.

set_defaults

# Give the window a title.

set title [label .title -text "Caltech Continuum Backend demonstration client" -bg black -fg yellow]

# Create the various display areas.

set phase [scan_cnf_phase_frame .phase]
set cal [scan_cnf_cal_frame .cal]
set timing [scan_cnf_timing_frame .timing]
set sampler [scan_cnf_sampler_frame .sampler]
set scan [scan_cmd_frame .scan]
set dump [dump_cmd_frame .dump]
set link [link_cmd_frame .link]
set tel [telemetry_cmd_frame .tel]
set sys [system_cmd_frame .sys]
set log [log_display_frame .log]
set integ [integ_display_frame .integ]
set monitor [monitor_display_frame .monitor]

# Stack the display areas vertically.

pack $title $phase $cal $timing $sampler $scan $dump $link $sys $tel $log $integ $monitor -side top -fill x -anchor center

# Register the callback procedures that receive notification of various
# CCB events.

ccb attach log ccb_log_callback
ccb attach status ccb_status_callback
ccb attach monitor ccb_monitor_callback
ccb attach integ ccb_integ_callback
ccb attach error ccb_error_callback

# Start the 1-second clock update procedure.

redisplay_time

# Start with the telemetry "off" button highlighted.

send_off_streams
