/*
 * Standard C-library includes.
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

/*
 * Local includes.
 */
#include "ccbserverlink.h"
#include "ccbdemodriver.h"
#include "ccbrealdriver.h"

/*
 * Encapsulate the running parameters of the demo program.
 */
typedef struct {
  CCBServerLink *link;          /* The communications interface of the server */
  CCBDemoDriverData *demo_data; /* The demo-driver's resources */
  CCBServerDriver *demo_driver; /* The demo-driver's interface object */
  CCBRealDriverData *real_data; /* The resources of the real CCB driver */
  CCBServerDriver *real_driver; /* The real driver's interface object */
} CCBServer;

static CCBServer *new_CCBServer(void);
static CCBServer *del_CCBServer(CCBServer *cs);

/*.......................................................................
 * This is the entry-point of the CCB server program.
 *
 * Arguments: none
 */
int main(int argc, char *argv[])
{
  CCBServer *cs;  /* The resources of the CCB server */
/*
 * Report when the server actually starts.
 */
  printf("The CCB server is now starting.\n");
/*
 * Establish the server.
 */
  cs = new_CCBServer();
  if(!cs)
    return 1;
/*
 * Run the server loop.
 */
  (void) ccb_server_event_loop(cs->link);
/*
 * Shutdown the server and release its resources.
 */
  cs = del_CCBServer(cs);
/*
 * Report termination.
 */
  printf("The CCB server is now terminating.\n");
  return 0;
}

/*.......................................................................
 * Create a new CCBServer object. This contains all of the resources
 * needed by the CCB server.
 *
 * Output:
 *  return  CCBServer *  The new object, or NULL on error.
 */
static CCBServer *new_CCBServer(void)
{
  CCBServer *cs;  /* The object to be returned */
/*
 * Allocate the container.
 */
  cs = malloc(sizeof(CCBServer));
  if(!cs) {
    fprintf(stderr, "new_CCBServer: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBServer().
 */
  cs->link = NULL;
  cs->demo_data = NULL;
  cs->demo_driver = NULL;
  cs->real_data = NULL;
  cs->real_driver = NULL;
/*
 * Allocate the resources of the demo-driver.
 */
  cs->demo_data = new_CCBDemoDriverData();
  if(!cs->demo_data)
    return del_CCBServer(cs);
/*
 * Allocate the interface object of the demo-driver.
 */
  cs->demo_driver = new_CCBDemoDriver(cs->demo_data);
  if(!cs->demo_driver)
    return del_CCBServer(cs);
/*
 * Allocate the resources of the real driver.
 */
  cs->real_data = new_CCBRealDriverData();
  if(!cs->real_data)
    return del_CCBServer(cs);
/*
 * Allocate the interface object of the real driver.
 */
  cs->real_driver = new_CCBRealDriver(cs->real_data);
  if(!cs->real_driver)
    return del_CCBServer(cs);
/*
 * Create the communications interface of the server.
 */
  cs->link = new_CCBServerLink(cs->real_driver, cs->demo_driver);
  if(!cs->link)
    return del_CCBServer(cs);
  return cs;
}

/*.......................................................................
 * Delete a CCBServer object.
 *
 * Input:
 *  cs     CCBServer *  The object to be deleted.
 * Output:
 *  return  CCBServer *  The deleted object (always NULL).
 */
static CCBServer *del_CCBServer(CCBServer *cs)
{
  if(cs) {
/*
 * Delete the communications interface.
 */
    cs->link = del_CCBServerLink(cs->link);
/*
 * Delete the interface objects of the drivers.
 */
    cs->demo_driver = del_CCBServerDriver(cs->demo_driver);
    cs->real_driver = del_CCBServerDriver(cs->real_driver);
/*
 * Delete the resource data of the drivers.
 */
    cs->demo_data = del_CCBDemoDriverData(cs->demo_data);
    cs->real_data = del_CCBRealDriverData(cs->real_data);
/*
 * Delete the container of the object.
 */
    free(cs);
  };
  return NULL;
}

