/*
 * This file provides a TCL interface to the CCB client-communications
 * library.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <tcl.h>

#include "ccbclientlink.h"
#include "ccbtclclient.h"

/*
 * Specify the size of the work-buffer in which TCL strings are composed.
 */
#define CTC_BUF_SIZE 256

/*
 * Callback scripts are aggregated in a structure of the following type.
 */
typedef struct {
  char *status;  /* The script that announces incoming status messages */
  char *monitor; /* The script that announces incoming monitor messages */
  char *integ;   /* The script that announces incoming integration messages */
  char *log;     /* The script that announces incoming log messages */
  char *error;   /* The script that announces incoming error messages */
} CCBTclScripts;

/*
 * The following constant is used as an initializer for CCBTclScripts
 * containers. It takes advantage of the fact that static data structures
 * are defined to be initialized as though 0 had been explicitly assigned
 * to each member of the structure.
 */
static const CCBTclScripts ccb_tcl_scripts_init;

/*
 * A structure of the following type is used to keep a copy of the
 * contents of the most recently received monitor-data message.
 */
typedef struct {
  CCBTimeStamp ts;      /* The time at which the message was sent */
  unsigned long scan;   /* The identifier of the parent scan */
  unsigned long number; /* The sequence number of the monitor data */
  unsigned long values[CCB_MAX_MONITOR]; /* The monitored values */
  unsigned nvalues;     /* The number of values in values[] */
} CtcMonitorMsg;

/*
 * A structure of the following type is used to keep a copy of the
 * contents of the most recently received integ-data message.
 */
typedef struct {
  CCBTimeStamp ts;      /* The time at which the message was sent */
  unsigned long scan;   /* The identifier of the parent scan */
  unsigned long number; /* The sequence number of the integration */
  unsigned short flags; /* Integration informational status-flags */
  unsigned long values[CCB_MAX_INTEG]; /* The integrated values */
  unsigned nvalues;     /* The number of values in values[] */
} CtcIntegMsg;

/*
 * A structure of the following type is used to keep a copy of the
 * contents of the most recently received log message.
 */
typedef struct {
  CCBTimeStamp ts;        /* The time at which the message was sent */
  char text[CCB_MAX_LOG]; /* The last log message */
  unsigned long id;       /* The ID of the originating log statement */
  CCBLogLevel level;      /* The significance level of the log message */
} CtcLogMsg;

/*
 * Information from recent messages received from the CCB server is
 * recorded in a container of the following type.
 */
typedef struct {
  unsigned long status;   /* The status of the CCB (a bitwise union of */
                          /*  CCBGeneralStatus enumerators). */
  CCBCmdStatus cmderror;  /* A command error status */
  CtcMonitorMsg monitor;  /* The last monitor data message */
  CtcIntegMsg integ;      /* The last integrated data message */
  CtcLogMsg log;          /* The last log message received */
} CCBTclMessages;

/*
 * The following constant is used as an initializer for CCBTclMessages
 * containers. It takes advantage of the fact that static data structures
 * are defined to be initialized as though 0 had been explicitly assigned
 * to each member of the structure.
 */
static const CCBTclMessages ccb_tcl_messages_init;

/*
 * Define the resource object of the Tcl interface.
 */
typedef struct {
  Tcl_Interp *interp;       /* The tcl-interpretter */
  CCBClientLink *client;    /* The CCB client communications object */
  CCBConfig *cnf;           /* The latest scan configuration */
  int cntrl_sock;           /* The file descriptor of the control socket */
  int telem_sock;           /* The file descriptor of the telemetry socket */
  char buffer[CTC_BUF_SIZE];/* A work buffer for constructing result strings */
  unsigned long cmd_id;     /* The sequence number of the last command that */
                            /*  was sent. */
  unsigned long scan_id;    /* The sequence number of the last scan that */
                            /*  was sent. */
  CCBTclScripts scripts;    /* Tcl callback scripts */
  CCBTclMessages messages;  /* Information from recently received messages */
} CCBTclClient;

static CCBTclClient *new_CCBTclClient(Tcl_Interp *interp);
static CCBTclClient *del_CCBTclClient(CCBTclClient *ctc);
static void delete_CCBTclClient(ClientData context);

/*
 * The function which implements the Tcl "ccb" command.
 */
static int ccb_tcl_command(ClientData context, Tcl_Interp *interp,
			   int argc, char *argv[]);

/*
 * The function that is called by the CCB library whenever sockets change
 * or whenever the I/O events needed by these sockets change.
 */
static CCB_CLIENT_SOCKETS_FN(ctc_client_sockets_fn);

/*
 * Functions which are called upon by ccb_tcl_command() to execute the
 * specifics of a given command are all defined and prototyped using
 * the following macro. They all return either TCL_OK or TCL_ERROR.
 */
#define CTC_SERVICE_FN(fn) int (fn)(CCBTclClient *ctc, int argc, char *argv[])

typedef CTC_SERVICE_FN(CTCServiceFn);

/*
 * Establish a new CCB connection.
 */
static CTC_SERVICE_FN(ctc_connect);

/*
 * Terminate the connection to the current CCB server.
 */
static CTC_SERVICE_FN(ctc_disconnect);

/*
 * Prototype the functions which send messages to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send);
static CTC_SERVICE_FN(ctc_send_start_scan);
static CTC_SERVICE_FN(ctc_send_stop_scan);
static CTC_SERVICE_FN(ctc_send_dump_scan);
static CTC_SERVICE_FN(ctc_send_load_driver);
static CTC_SERVICE_FN(ctc_send_reset);
static CTC_SERVICE_FN(ctc_send_ping);
static CTC_SERVICE_FN(ctc_send_status_request);
static CTC_SERVICE_FN(ctc_send_shutdown);
static CTC_SERVICE_FN(ctc_send_reboot);
static CTC_SERVICE_FN(ctc_send_monitor);
static CTC_SERVICE_FN(ctc_send_telemetry);
static CTC_SERVICE_FN(ctc_send_logger);

/*
 * Prototype the functions which configure scan parameters.
 */
static CTC_SERVICE_FN(ctc_configure);
static CTC_SERVICE_FN(ctc_conf_phase_switches);
static CTC_SERVICE_FN(ctc_conf_cal_diode);
static CTC_SERVICE_FN(ctc_conf_timing);
static CTC_SERVICE_FN(ctc_conf_sampler);

/*
 * Prototype the function that attach Tcl callback scripts to the receipt
 * of incoming messages.
 */
static CTC_SERVICE_FN(ctc_attach);

/*
 * Prototype the functions that return information received from the CCB
 * server.
 */
static CTC_SERVICE_FN(ctc_get);
static int ctc_get_monitor(CCBTclClient *ctc);
static int ctc_get_integ(CCBTclClient *ctc);
static int ctc_get_log(CCBTclClient *ctc);
static int ctc_get_status(CCBTclClient *ctc);
static int ctc_get_error(CCBTclClient *ctc);
static int ctc_get_time(CCBTclClient *ctc);

/*
 * The following are the prototypes of the CCB callback functions
 * that deliver received messages.
 */
static CCB_CMD_ERROR_FN(ctc_cmd_error_fn);
static CCB_STATUS_REPLY_FN(ctc_status_reply_fn);
static CCB_MONITOR_MSG_FN(ctc_monitor_msg_fn);
static CCB_INTEG_MSG_FN(ctc_integ_msg_fn);
static CCB_LOG_MSG_FN(ctc_log_msg_fn);

/*
 * Parse the contents of a Tcl string argument as an unsigned integer.
 */
static int ctc_get_unsigned_scalar(CCBTclClient *ctc, const char *name,
				   const char *string,
				   unsigned long *ulong_var,
				   unsigned short *ushort_var,
				   unsigned char *uchar_var);

/*
 * Parse the contents of a Tcl string argument as an unsigned integer array.
 */
static int ctc_get_unsigned_array(CCBTclClient *ctc, const char *name,
				  const char *string, int nmin, int nmax,
				  int step, unsigned long *ulongs,
				  unsigned short *ushorts,
				  unsigned char *uchars, unsigned *nval);

/*
 * Provide a means to associate a name with an unsigned short value.
 */
typedef struct {
  const char *name;     /* The name of the enumerator */
  unsigned short value; /* The value of the enumerator */
} CtcUShortSymbol;

static int ctc_get_ushort_bitmask(CCBTclClient *ctc, const char *command,
				  const char *type_name, const char *string,
				  const CtcUShortSymbol *sym, int nsym,
				  unsigned short *mask);

/*
 * Define a table associating names with enumerated CCBTelemetryStream
 * enumerators.
 */
static const CtcUShortSymbol ctc_stream_symbol_table[] = {
  {"integ_stream", CCB_INTEG_STREAM}, {"monitor_stream", CCB_MONITOR_STREAM},
  {"log_stream", CCB_LOG_STREAM}, {"no_streams", CCB_NO_STREAMS},
  {"all_streams", CCB_ALL_STREAMS}
};
static const int ctc_stream_symbol_count = sizeof(ctc_stream_symbol_table) /
                                           sizeof(ctc_stream_symbol_table[0]);

/*
 * Parse a space-separated list of stream names, returning the corresponding
 * bitset union of CCBTelemetryStream enumerators.
 */
static int ctc_get_stream_mask(CCBTclClient *ctc, const char *command,
			       const char *string, unsigned short *streams);

/*
 * Define a table associating names with enumerated CCBCalDiodes
 * enumerators.
 */
static const CtcUShortSymbol ctc_diode_symbol_table[] = {
  {"diode_a", CCB_CAL_DIODE_A}, {"diode_b", CCB_CAL_DIODE_B},
  {"no_diodes", CCB_NO_CAL_DIODES},
  {"all_diodes", CCB_ALL_CAL_DIODES}
};
static const int ctc_diode_symbol_count = sizeof(ctc_diode_symbol_table) /
                                          sizeof(ctc_diode_symbol_table[0]);

/*
 * Parse a space-separated list of cal-diode names, returning the
 * corresponding bitset union of CCBCalDiodes enumerators.
 */
static int ctc_get_cal_diode_mask(CCBTclClient *ctc, const char *command,
				  const char *string, unsigned short *diodes);

/*
 * Define a table associating names with enumerated CCBPhaseSwitches
 * enumerators.
 */
static const CtcUShortSymbol ctc_switch_symbol_table[] = {
  {"switch_a", CCB_PHASE_SWITCH_A}, {"switch_b", CCB_PHASE_SWITCH_B},
  {"no_switches", CCB_NO_PHASE_SWITCHES},
  {"all_switches", CCB_ALL_PHASE_SWITCHES}
};
static const int ctc_switch_symbol_count = sizeof(ctc_switch_symbol_table) /
                                           sizeof(ctc_switch_symbol_table[0]);

/*
 * Parse a space-separated list of phase-switch names, returning the
 * corresponding bitset union of CCBPhaseSwitches enumerators.
 */
static int ctc_get_phase_switch_mask(CCBTclClient *ctc, const char *command,
				     const char *string,
				     unsigned short *switches);

/*.......................................................................
 * This is the TCL monitor-viewer package-initialization function.
 *
 * Input:
 *  interp  Tcl_Interp *  The TCL interpretter.
 * Output:
 *  return         int    TCL_OK    - Success.
 *                        TCL_ERROR - Failure.           
 */
int Ccbtclclient_Init(Tcl_Interp *interp)
{
  CCBTclClient *ctc;   /* The resource object of the tcl_db() function */
/*
 * Check arguments.
 */
  if(!interp) {
    fprintf(stderr, "Ccbtclclient_Init: NULL interpreter.\n");
    return TCL_ERROR;
  };
/*
 * Allocate a resource object for the service_CCBTclClient() command.
 */
  ctc = new_CCBTclClient(interp);
  if(!ctc)
    return TCL_ERROR;
/*
 * Create the Tcl commands that implement the Tcl interface to the
 * CCBClientLink object in ctc.
 */
  if(Tcl_CreateCommand(interp, (char *)"ccb", ccb_tcl_command,
		       (ClientData *)ctc, delete_CCBTclClient) == NULL) {
    ctc = del_CCBTclClient(ctc);
    return TCL_ERROR;
  };
  return TCL_OK;
}

/*.......................................................................
 * Create the resource object of the CCB client TCL interface.
 *
 * Input:
 *  interp  Tcl_Interp *  The Tcl interpretter.
 * Output:
 *  return  CCBTclClient *  The resource object, or NULL on error.
 */
static CCBTclClient *new_CCBTclClient(Tcl_Interp *interp)
{
  CCBTclClient *ctc;  /* The object to be returned */
/*
 * Check arguments.
 */
  if(!interp) {
    fprintf(stderr, "new_CCBTclClient: NULL interp argument.\n");
    return NULL;
  };
/*
 * Allocate the container.
 */
  ctc = (CCBTclClient *) malloc(sizeof(CCBTclClient));
  if(!ctc) {
    fprintf(stderr, "new_CCBTclClient: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBTclClient().
 */
  ctc->interp = interp;
  ctc->client = NULL;
  ctc->cnf = NULL;
  ctc->cntrl_sock = -1;
  ctc->telem_sock = -1;
  ctc->cmd_id = 0;
  ctc->scan_id = 0;
  ctc->scripts = ccb_tcl_scripts_init;
  ctc->messages = ccb_tcl_messages_init;
/*
 * Create the CCB client communications object.
 */
  ctc->client = new_CCBClientLink(CTC_BUF_SIZE, ctc->buffer);
  if(!ctc->client) {
    fprintf(stderr, "%s\n", ctc->buffer);
    return del_CCBTclClient(ctc);
  };
/*
 * Create a CCB configuration object.
 */
  ctc->cnf = new_CCBConfig();
  if(!ctc->cnf) {
    fprintf(stderr, "Error allocating configuration object (%s).\n",
	    strerror(errno));
    return del_CCBTclClient(ctc);
  };
/*
 * Tell the library which function to call when it needs to inform the
 * Tcl event loop of changes to the events and/or file descriptors that
 * it should be watching.
 */
  if(ccb_client_sockets_callback(ctc->client, ctc_client_sockets_fn, ctc))
    return del_CCBTclClient(ctc);
/*
 * Register the callback functions that deliver messages to us.
 */
  if(ccb_cmd_error_callback(ctc->client, ctc_cmd_error_fn, ctc) ||
     ccb_status_reply_callback(ctc->client, ctc_status_reply_fn,
				     ctc) ||
     ccb_monitor_msg_callback(ctc->client, ctc_monitor_msg_fn, ctc) ||
     ccb_integ_msg_callback(ctc->client, ctc_integ_msg_fn, ctc) ||
     ccb_log_msg_callback(ctc->client, ctc_log_msg_fn, ctc))
    return del_CCBTclClient(ctc);
  return ctc;
}

/*.......................................................................
 * Delete the resource object of the CCB client TCL interface.
 *
 * Input:
 *  ctc    CCBTclClient *  The object to be deleted.
 * Output:
 *  return CCBTclClient *  The deleted object (Always NULL).
 */
static CCBTclClient *del_CCBTclClient(CCBTclClient *ctc)
{
  if(ctc) {
/*
 * Delete the CCB communications object.
 */
    ctc->client = del_CCBClientLink(ctc->client);
/*
 * Delete the CCB configuration object.
 */
    ctc->cnf = del_CCBConfig(ctc->cnf);
/*
 * Delete the container.
 */
    free(ctc);
  };
  return NULL;
}

/*.......................................................................
 * This is a wrapper around del_CCBTclClient() that can be called with a
 * ClientData version of the (CCBTclClient *) pointer.
 *
 * Input:
 *  context    ClientData  A CCBTclClient pointer cast to ClientData.
 */
static void delete_CCBTclClient(ClientData context)
{
  del_CCBTclClient((CCBTclClient *)context);
}

/*.......................................................................
 * This is the service function of the Tcl monitor command. It dispatches
 * the appropriate function for the service that is named by the
 * first argument.
 *
 * Input:
 *  context   ClientData    The CCBTclClient resource object cast to
 *                          ClientData.
 *  interp    Tcl_Interp *  The Tcl interpretter instance.
 *  argc             int    The number of arguments in argv[].
 *  argv            char ** The arguments of the command. The first
 *                          argument specifies what service is required.
 * Output:
 *  return           int    TCL_OK    - Success.
 *                          TCL_ERROR - Error.
 */
static int ccb_tcl_command(ClientData context, Tcl_Interp *interp,
			   int argc, char *argv[])
{
  const char *service;   /* The name of the desired CCB service */
/*
 * Get our resource object from the callback data.
 */
  CCBTclClient *ctc = (CCBTclClient *) context;
/*
 * Check arguments.
 */
  if(!ctc || !interp || !argv) {
    fprintf(stderr, "ccb_tcl_command: NULL arguments.\n");
    return TCL_ERROR;
  };
  if(argc < 2) {
    Tcl_AppendResult(interp, "ccb command name missing.", NULL);
    return TCL_ERROR;
  };
/*
 * Get the name of the CCB service to invoke.
 */
  service = argv[1];
/*
 * Skip over the command and service names to the start of the
 * arguments of the service.
 */
  argv += 2;
  argc -= 2;
#ifdef DEBUG
  {
    int i;
    printf("ccb_tcl_command(service=\"%s\"", service);
    for(i=0; i<argc; i++)
      printf(", %s", argv[i]);
    printf(")\n");
  };
#endif
/*
 * Identify the service, then call the function that implements it.
 */
  if(strcmp(service, "connect") == 0)
    return ctc_connect(ctc, argc, argv);
  else if(strcmp(service, "disconnect") == 0)
    return ctc_disconnect(ctc, argc, argv);
  else if(strcmp(service, "send") == 0)
    return ctc_send(ctc, argc, argv);
  else if(strcmp(service, "configure") == 0)
    return ctc_configure(ctc, argc, argv);
  else if(strcmp(service, "attach") == 0)
    return ctc_attach(ctc, argc, argv);
  else if(strcmp(service, "get") == 0)
    return ctc_get(ctc, argc, argv);
/*
 * Unrecognized command name?
 */
  Tcl_AppendResult(interp, "Unknown CCB service: ", (char *)service, NULL);
  return TCL_ERROR;
}

/*.......................................................................
 * Implement the Tcl CCB service that establishes a new connection to a
 * CCB server.
 *
 * Input:
 *  ctc   CCBTclClient *  The resource object of the CCB Tcl interface.
 *  argc           int    The number of Tcl arguments in argv[].
 *  argv[]        char *  The array of Tcl arguments.
 * Output:
 *  return         int    TCL_OK     -  Success.
 *                        TCL_ERROR  -  Error. A message is left in the
 *                                      Tcl result string.
 */
static CTC_SERVICE_FN(ctc_connect)
{
/*
 * There should be a single argument.
 */
  if(argc != 1) {
    Tcl_AppendResult(ctc->interp, "ccb connect: Wrong number of arguments.",
		     NULL);
    return 1;
  };
/*
 * Attempt to initiate a non-blocking connection to the specified CCB
 * server.
 */
  if(ccb_client_connect(ctc->client, argv[0], 1))
    return TCL_ERROR;
/*
 * Reset the scan sequence number for the new connection.
 */
  ctc->scan_id = 0;
/*
 * Discard the most recent messages received from the previous connection.
 */
  ctc->messages = ccb_tcl_messages_init;
  return TCL_OK;
}

/*.......................................................................
 * Implement the Tcl CCB service that shuts down any existing connection
 * to a CCB server.
 *
 * Input:
 *  ctc   CCBTclClient *  The resource object of the CCB Tcl interface.
 *  argc           int    The number of Tcl arguments in argv[].
 *  argv[]        char *  The array of Tcl arguments.
 * Output:
 *  return         int    TCL_OK     -  Success.
 *                        TCL_ERROR  -  Error. A message is left in the
 *                                      Tcl result string.
 */
static CTC_SERVICE_FN(ctc_disconnect)
{
/*
 * There shouldn't be any arguments.
 */
  if(argc != 0) {
    Tcl_AppendResult(ctc->interp,
		     "ccb disconnect: Wrong number of arguments.", NULL);
    return 1;
  };
  ccb_client_disconnect(ctc->client);
  return TCL_OK;
}

/*.......................................................................
 * Implement the service which sends specified commands to a remote CCB
 * server.
 *
 * Input:
 *  ctc   CCBTclClient *  The resource object of the CCB Tcl interface.
 *  argc           int    The number of Tcl arguments in argv[].
 *  argv[]        char *  The array of Tcl arguments.
 * Output:
 *  return         int    TCL_OK     -  Success.
 *                        TCL_ERROR  -  Error. A message is left in the
 *                                      Tcl result string.
 */
static CTC_SERVICE_FN(ctc_send)
{
  char *type;   /* The type of message to be sent */
  int i;
/*
 * List the types of messages that can be sent, along with the number
 * of arguments that they require.
 */
  typedef struct {
    const char *type;   /* The name of the message-type */
    int min_args;       /* The minimum number of acceptable arguments */
    int max_args;       /* The maximum number of acceptable arguments */
    int step;           /* Between min_args and max_args, the acceptable */
                        /*  number of arguments are restricted to */
                        /*  (min_args + n*step), where n is a positive */
                        /*  integer. */
    CTCServiceFn *fn;   /* The function which sends the message */
  } CTCMessageType;
  static CTCMessageType message_table[] = {
    {"start_scan",     2, 2, 1, ctc_send_start_scan},
    {"stop_scan",      0, 0, 1, ctc_send_stop_scan},
    {"dump_scan",      3, 3, 1, ctc_send_dump_scan},
    {"load_driver",    1, 1, 1, ctc_send_load_driver},
    {"reset",          0, 0, 1, ctc_send_reset},
    {"ping",           0, 0, 1, ctc_send_ping},
    {"status_request", 0, 0, 1, ctc_send_status_request},
    {"shutdown",       0, 0, 1, ctc_send_shutdown},
    {"reboot",         0, 0, 1, ctc_send_reboot},
    {"monitor",        1, 1, 1, ctc_send_monitor},
    {"telemetry",      1, 1, 1, ctc_send_telemetry},
    {"logger",         1, 1, 1, ctc_send_logger},
  };
/*
 * At least one argument is needed to identify the type of message to
 * be sent.
 */
  if(argc < 1) {
    Tcl_AppendResult(ctc->interp, "ccb send: Missing message type.", NULL);
    return TCL_ERROR;
  };
/*
 * Get the name of the message type to be sent.
 */
  type = argv[0];
/*
 * Skip over the type argument to the arguments which denote the message
 * contents.
 */
  argv++;
  argc--;
/*
 * Identify the type of message to be sent.
 */
  for(i=0; i<sizeof(message_table)/sizeof(message_table[0]); i++) {
    CTCMessageType *mt = message_table + i;
/*
 * Does the type name match that of the latest table entry?
 */
    if(strcmp(type, mt->type) == 0) {
/*
 * Check that the correct number of arguments have been supplied.
 */
      if(argc < mt->min_args || argc > mt->max_args ||
	 (argc - mt->min_args) % mt->step != 0) {
	Tcl_AppendResult(ctc->interp, "ccb send ", (char *)type,
			 ": Wrong number of arguments.", NULL);
	return TCL_ERROR;
      };
/*
 * Invoke the function which sends the identified type of message.
 */
      return mt->fn(ctc, argc, argv);
    };
  };
/*
 * Message-type not recognized.
 */
  Tcl_AppendResult(ctc->interp, "ccb send: Message type (",
		   (char *)type, ") unknown.", NULL);
  return TCL_ERROR;
}

/*.......................................................................
 * Implement the service which configures specified configuration groups.
 *
 * Input:
 *  ctc   CCBTclClient *  The resource object of the CCB Tcl interface.
 *  argc           int    The number of Tcl arguments in argv[].
 *  argv[]        char *  The array of Tcl arguments.
 * Output:
 *  return         int    TCL_OK     -  Success.
 *                        TCL_ERROR  -  Error. A message is left in the
 *                                      Tcl result string.
 */
static CTC_SERVICE_FN(ctc_configure)
{
  char *type;   /* The type of configuration */
  int i;
/*
 * List the types of configuration, along with the number of arguments
 * that they require.
 */
  typedef struct {
    const char *type;   /* The name of the configuration-type */
    int min_args;       /* The minimum number of acceptable arguments */
    int max_args;       /* The maximum number of acceptable arguments */
    int step;           /* Between min_args and max_args, the acceptable */
                        /*  number of arguments are restricted to */
                        /*  (min_args + n*step), where n is a positive */
                        /*  integer. */
    CTCServiceFn *fn;   /* The function which applies the configuration */
  } CTCConfigType;
  static CTCConfigType config_table[] = {
    {"phase_switches",  3, 3, 1, ctc_conf_phase_switches},
    {"cal_diode",       3, 3, 1, ctc_conf_cal_diode},
    {"timing",          6, 6, 1, ctc_conf_timing},
    {"sampler",         1, 1, 1, ctc_conf_sampler},
  };
/*
 * At least one argument is needed to identify the type of
 * configuration.
 */
  if(argc < 1) {
    Tcl_AppendResult(ctc->interp, "ccb configure: Missing configuration type.",
		     NULL);
    return TCL_ERROR;
  };
/*
 * Get the name of the configuration type to be sent.
 */
  type = argv[0];
/*
 * Skip over the type argument to the arguments which denote the configuration
 * parameters.
 */
  argv++;
  argc--;
/*
 * Identify the type of configuration to be sent.
 */
  for(i=0; i<sizeof(config_table)/sizeof(config_table[0]); i++) {
    CTCConfigType *ct = config_table + i;
/*
 * Does the type name match that of the latest table entry?
 */
    if(strcmp(type, ct->type) == 0) {
/*
 * Check that the correct number of arguments have been supplied.
 */
      if(argc < ct->min_args || argc > ct->max_args ||
	 (argc - ct->min_args) % ct->step != 0) {
	Tcl_AppendResult(ctc->interp, "ccb configure ", (char *)type,
			 ": Wrong number of arguments.", NULL);
	return TCL_ERROR;
      };
/*
 * Invoke the function which sends the identified type of configuration.
 */
      return ct->fn(ctc, argc, argv);
    };
  };
/*
 * Configuration-type not recognized.
 */
  Tcl_AppendResult(ctc->interp, "ccb configure: Configuration type (",
		   (char *)type, ") unknown.", NULL);
  return TCL_ERROR;
}

/*.......................................................................
 * Implement the service which attaches callback functions to asynchronous
 * events.
 *
 * Input:
 *  ctc   CCBTclClient *  The resource object of the CCB Tcl interface.
 *  argc           int    The number of Tcl arguments in argv[].
 *  argv[]        char *  The array of Tcl arguments.
 * Output:
 *  return         int    TCL_OK     -  Success.
 *                        TCL_ERROR  -  Error. A message is left in the
 *                                      Tcl result string.
 */
static CTC_SERVICE_FN(ctc_attach)
{
  const char *type;          /* The type of callback */
  const char *script = NULL; /* The callback script */
  char **field;              /* A pointer to the field that records the */
                             /*  callback. */
/*
 * Get the type of the callback, and the callback script itself.
 */
  switch(argc) {
  case 2:
    script = argv[1];
    /* Note fallthrough to case 1 */
  case 1:
    type = argv[0];
    break;
  case 0:
    Tcl_AppendResult(ctc->interp, "ccb attach: Missing callback type.", NULL);
    return TCL_ERROR;
    break;
  default:
    Tcl_AppendResult(ctc->interp, "ccb attach: Wrong number of arguments",
		     NULL);
    return TCL_ERROR;
  };
/*
 * Identify the type of callback being registered, and get a pointer to
 * the corresponding field in ctc.
 */
  if(strcmp(type, "status") == 0) {
    field = &ctc->scripts.status;
  } else if(strcmp(type, "monitor") == 0) {
    field = &ctc->scripts.monitor;
  } else if(strcmp(type, "integ") == 0) {
    field = &ctc->scripts.integ;
  } else if(strcmp(type, "log") == 0) {
    field = &ctc->scripts.log;
  } else if(strcmp(type, "error") == 0) {
    field = &ctc->scripts.error;
  } else {
    Tcl_AppendResult(ctc->interp, "ccb attach: Callback type (", (char *)type,
		     ") unknown.", NULL);
    return TCL_ERROR;
  };
/*
 * Delete any existing callback.
 */
  if(*field) {
    free(*field);
    *field = NULL;
  };
/*
 * If a new callback script has been provided, allocate memory for a copy
 * of it.
 */
  if(script && *script) {
    *field = malloc(strlen(script) + 1);
    if(!*field) {
      Tcl_AppendResult(ctc->interp,
		       "Insufficient memory to record callback script.", NULL);
      return TCL_ERROR;
    };
/*
 * Make a copy of the callback script.
 */
    strcpy(*field, script);
  };
  return TCL_OK;
}

/*.......................................................................
 * Implement the service which returns the contents of the most recent of
 * a specified type of message, received from the CCB server.
 *
 * Input:
 *  ctc   CCBTclClient *  The resource object of the CCB Tcl interface.
 *  argc           int    The number of Tcl arguments in argv[].
 *  argv[]        char *  The array of Tcl arguments.
 * Output:
 *  return         int    TCL_OK     -  Success. The result string will
 *                                      contain the message contents.
 *                        TCL_ERROR  -  Error. A message is left in the
 *                                      Tcl result string.
 */
static CTC_SERVICE_FN(ctc_get)
{
  const char *type;   /* The type of message callback */
/*
 * At least one argument is needed to identify the information to be
 * returned.
 */
  if(argc < 1) {
    Tcl_AppendResult(ctc->interp, "ccb get: Missing information type.",
		     NULL);
    return TCL_ERROR;
  };
/*
 * Get the type of the message.
 */
  type = argv[0];
/*
 * Identify the type of message that is wanted, and call the function that
 * returns it.
 */
  if(strcmp(type, "status") == 0) {
    return ctc_get_status(ctc);
  } else if(strcmp(type, "monitor") == 0) {
    return ctc_get_monitor(ctc);
  } else if(strcmp(type, "integ") == 0) {
    return ctc_get_integ(ctc);
  } else if(strcmp(type, "log") == 0) {
    return ctc_get_log(ctc);
  } else if(strcmp(type, "error") == 0) {
    return ctc_get_error(ctc);
  } else if(strcmp(type, "time") == 0) {
    return ctc_get_time(ctc);
  };
  Tcl_AppendResult(ctc->interp, "ccb get: Requested information (",
		   (char *) type, ") unknown.", NULL);
  return TCL_ERROR;
}

/*.......................................................................
 * Read an unsigned integer from a Tcl argument string.
 *
 * Input:
 *  ctc           CCBTclClient *  The resource object of the CCB Tcl
 *                                interface.
 *  name            const char *  The name of the target item, for use
 *                                in error messages.
 *  string          const char *  The Tcl argument string to read from.
 * Input:
 *  ulong_var    unsigned long *  If a pointer to a variable is passed
 *                                in this argument, the values read
 *                                will be recorded in it as unsigned
 *                                long elements. If not required, pass
 *                                NULL.
 *  ushort_var  unsigned short *  If a pointer to a variable is passed
 *                                in this argument, the values read
 *                                will be recorded in it as unsigned
 *                                short elements. If not required,
 *                                pass NULL.
 *  uchar_var    unsigned char *  If a pointer to a variable is passed
 *                                in this argument, the values read
 *                                will be recorded in it as unsigned
 *                                char elements. If not required,
 *                                pass NULL.
 * Output:
 *  return         int    TCL_OK    - OK.
 *                        TCL_ERROR - Error (An error message is placed
 *                                    in the Tcl result string).
 */
static int ctc_get_unsigned_scalar(CCBTclClient *ctc, const char *name,
				   const char *string,
				   unsigned long *ulong_var,
				   unsigned short *ushort_var,
				   unsigned char *uchar_var)
{
  char *endp;  /* A pointer to the character in string[] that follows the */
               /*  integer. */
/*
 * Attempt to convert the string to a number.
 */
  unsigned long number = strtoul(string, &endp, 0);
/*
 * Check that the argument was converted successfully.
 */
  if(*string == '\0' || *endp != '\0') {
    Tcl_AppendResult(ctc->interp, "Error reading ", (char *)name, ".", NULL);
    return TCL_ERROR;
  };
/*
 * Return the new number.
 */
  if(ulong_var)
    *ulong_var = number;
  if(ushort_var) {
    *ushort_var = number;
    if((unsigned long) *ushort_var != number) {
      Tcl_AppendResult(ctc->interp, (char *)name, " too big\n", NULL);
      return TCL_ERROR;
    };
  };
  if(uchar_var) {
    *uchar_var = number;
    if((unsigned long) *uchar_var != number) {
      Tcl_AppendResult(ctc->interp, (char *)name, " too big\n", NULL);
      return TCL_ERROR;
    };
  };
  return TCL_OK;
}

/*.......................................................................
 * Send a start-scan command to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_start_scan)
{
  unsigned long mjd, tod;   /* The target date and time-of-day */
  int arg=0;                /* The index of the argument being processed */
/*
 * Get the start date and time.
 */
  if(ctc_get_unsigned_scalar(ctc, "start_scan MJD", argv[arg++], &mjd,
			     NULL, NULL) ||
     ctc_get_unsigned_scalar(ctc, "start_scan tod", argv[arg++], &tod,
			     NULL, NULL))
    return TCL_ERROR;
/*
 * Send the command.
 */
  if(ccb_queue_start_scan_cmd(ctc->client, ++ctc->cmd_id, ctc->cnf, ++ctc->scan_id,
			      mjd, tod))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Send a stop-scan command to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_stop_scan)
{
  if(ccb_queue_stop_scan_cmd(ctc->client, ++ctc->cmd_id, ctc->cnf,
			     ++ctc->scan_id))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Send a dump-scan command to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_dump_scan)
{
  unsigned short adc;        /* The ADC whose samples are to be collected */
  unsigned long samples;     /* The maximum number of samples to collect */
                             /*  per integration (or 0 to select max). */
  unsigned long frames;      /* The number of per-integration data-frames */
                             /*  to deliver to the named pipe (or 0 to */
                             /*  deliver frames indefinitely). */
  int arg=0;                 /* The index of the argument being processed */
/*
 * Get the start date and time.
 */
  if(ctc_get_unsigned_scalar(ctc, "dump_scan adc", argv[arg++], NULL, &adc,
			     NULL))
    return TCL_ERROR;
/*
 * The 'samples' parameter can be either the word "max", or a positive integer.
 */
  if(strcmp(argv[arg], "max")==0) {
    samples = CCB_DUMP_MAX_SAMPLES;
    arg++;
  } else if(ctc_get_unsigned_scalar(ctc, "dump_scan samples",
				    argv[arg++], &samples, NULL, NULL)) {
    return TCL_ERROR;
  };
/*
 * The 'frames' parameter can be either the word "all", or a positive integer.
 */
  if(strcmp(argv[arg], "all")==0) {
    frames = CCB_DUMP_ALL_FRAMES;
    arg++;
  } else if(ctc_get_unsigned_scalar(ctc, "dump_scan frames",
				    argv[arg++], &frames, NULL, NULL)) {
    return TCL_ERROR;
  };
/*
 * Queue the command for dispatch to the server.
 */
  if(ccb_queue_dump_scan_cmd(ctc->client, ++ctc->cmd_id, ctc->cnf,
			     ++ctc->scan_id, adc, samples, frames))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Send a load-driver command to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_load_driver)
{
  unsigned short type;    /* The type of driver to load */
  char *name;             /* The textual name of the type of driver to load */
/*
 * The 'type' parameter should name a member of the CCBDriverType enumeration.
 */
  name = argv[0];
  if(strcmp(name, "normal")==0) {
    type = CCB_NORMAL_DRIVER;
  } else if(strcmp(name, "virtual")==0) {
    type = CCB_VIRTUAL_DRIVER;
  } else {
    Tcl_AppendResult(ctc->interp, "Unknown driver type: \"", name, "\"\n",NULL);
    return TCL_ERROR;
  };
/*
 * Queue the command for dispatch to the server.
 */
  if(ccb_queue_load_driver_cmd(ctc->client, ++ctc->cmd_id, type))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Send a reset command to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_reset)
{
  if(ccb_queue_reset_cmd(ctc->client, ++ctc->cmd_id))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Send a ping command to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_ping)
{
  int status = TCL_OK;   /* The return status */
/*
 * See if the previous ping command has been replied to.
 */
  if(ccb_ping_echos(ctc->client) != CCB_ALL_LINKS) {
    Tcl_AppendResult(ctc->interp,
         "No response has been received to the previous ping command.", NULL);
    status = TCL_ERROR;
  };
/*
 * Send the new ping command.
 */
  if(ccb_queue_ping_cmd(ctc->client, ++ctc->cmd_id))
    return TCL_ERROR;
  return status;
}

/*.......................................................................
 * Send a status-request command message to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_status_request)
{
  if(ccb_queue_status_request_cmd(ctc->client, ++ctc->cmd_id))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Send a shutdown message to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_shutdown)
{
  if(ccb_queue_shutdown_cmd(ctc->client, ++ctc->cmd_id))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Send a reboot message to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_reboot)
{
  if(ccb_queue_reboot_cmd(ctc->client, ++ctc->cmd_id))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Send a monitor-configuration command message to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_monitor)
{
  unsigned short period;  /* The monitoring period */
  int arg=0;              /* The index of the argument being processed */
/*
 * Get the monitoring period.
 */
  if(ctc_get_unsigned_scalar(ctc, "monitor period", argv[arg++],
			     NULL, &period, NULL))
    return TCL_ERROR;
/*
 * Send the command.
 */
  if(ccb_queue_monitor_cmd(ctc->client, ++ctc->cmd_id, period))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Send a telemetry configuration command to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_telemetry)
{
  unsigned short streams;  /* The set of telemetry streams to send. */
  int arg=0;               /* The index of the argument being processed */
/*
 * Get the stream mask.
 */
  if(ctc_get_stream_mask(ctc, "ccb send telemetry", argv[arg++], &streams))
    return TCL_ERROR;
/*
 * Send the command.
 */
  if(ccb_queue_telemetry_cmd(ctc->client, ++ctc->cmd_id, streams))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Send a logger configuration command to the CCB server.
 */
static CTC_SERVICE_FN(ctc_send_logger)
{
  unsigned short period;  /* The logger historical-message-purging period */
  int arg=0;              /* The index of the argument being processed */
/*
 * Get the historical log purging period.
 */
  if(ctc_get_unsigned_scalar(ctc, "monitor period", argv[arg++],
			     NULL, &period, NULL))
    return TCL_ERROR;
/*
 * Send the command.
 */
  if(ccb_queue_logger_cmd(ctc->client, ++ctc->cmd_id, period))
    return TCL_ERROR;
  return TCL_OK;
}

/*.......................................................................
 * Install a new phase-switch configuration for the next scan.
 */
static CTC_SERVICE_FN(ctc_conf_phase_switches)
{
  CCBPhaseSwitchCnf phase;  /* The new configuration */
  int arg=0;                /* The index of the argument being processed */
/*
 * Get the configuration arguments.
 */
  if(ctc_get_phase_switch_mask(ctc, "ccb configure phase_switch", argv[arg++],
			       &phase.active_switches) ||
     ctc_get_phase_switch_mask(ctc, "ccb configure phase_switch", argv[arg++],
			       &phase.closed_switches) ||
     ctc_get_unsigned_scalar(ctc, "phase_switch samp_per_state", argv[arg++],
			     NULL, &phase.samp_per_state, NULL))
    return TCL_ERROR;
/*
 * Install the new configuration.
 */
  if(ccb_set_phase_switch_cnf(ctc->cnf, phase.active_switches,
			      phase.closed_switches, phase.samp_per_state)) {
    Tcl_AppendResult(ctc->interp, "ccb phase_switches: ",
		     strerror(errno), ".", NULL);
    return TCL_ERROR;
  };
  return TCL_OK;
}

/*.......................................................................
 * Install a new calibration-diode configuration for the next scan.
 */
static CTC_SERVICE_FN(ctc_conf_cal_diode)
{
  CCBCalDiodeCnf cal;  /* The cal-diode configuration */
  int arg=0;           /* The index of the argument being processed */
  int nstate;          /* The number of elements in the list of cal-diode */
                       /*  states. */
  char **states;       /* The array of cal-diode state-specifications */
  int waserr = 0;      /* True if an error occurs while there is memory */
                       /*  to be deallocated. */
  int i;
/*
 * Get the number of calibration steps.
 */
  if(ctc_get_unsigned_scalar(ctc, "cal_diode ncal", argv[arg++],
			     NULL, &cal.ncal, NULL))
    return TCL_ERROR;
/*
 * Check that the requested number of calibration steps can be accomodated.
 * in the state-machine configuration-array arguments.
 */
  if(cal.ncal > CCB_MAX_NCAL) {
    Tcl_AppendResult(ctc->interp, "Too many calibration steps requested.",
		     NULL);
    return TCL_ERROR;
  };
/*
 * Split the list of calibration diode state-specifications into an
 * array of its members.
 */
  if(Tcl_SplitList(ctc->interp, argv[arg++], &nstate, &states))
    return TCL_ERROR;
/*
 * Check that there are at least as many states listed as the specified
 * number of steps.
 */
  if(nstate < cal.ncal) {
    Tcl_AppendResult(ctc->interp,
	     "ccb configure cal_diode: Too few calibration-diode states.",
	     NULL);
    waserr = 1;
  };
/*
 * Get the array of cal-diode states.
 */
  for(i=0; !waserr && i<cal.ncal; i++) {
    waserr = ctc_get_cal_diode_mask(ctc, "ccb configure cal_diode",states[i],
				    &cal.diode_states[i]);
  };
/*
 * Discard the list of cal-diode specifications.
 */
  Tcl_Free((char *)states);
/*
 * If an error was postponed until we had freed the states[] array,
 * act on it now.
 */
  if(waserr)
    return TCL_ERROR;
/*
 * Get the array of calibration-diode state durations.
 */
  if(ctc_get_unsigned_array(ctc, "cal_diode diode_times[]", argv[arg++],
			    cal.ncal, CCB_MAX_NCAL, 1, cal.diode_times,
			    NULL, NULL, NULL))
    return TCL_ERROR;
/*
 * Install the new configuration.
 */
  if(ccb_set_cal_diode_cnf(ctc->cnf, cal.ncal, cal.diode_states,
			   cal.diode_times)) {
    Tcl_AppendResult(ctc->interp, "ccb cal_diode: ", strerror(errno), ".",
		     NULL);
    return TCL_ERROR;
  };
  return TCL_OK;
}

/*.......................................................................
 * Install a new timing configuration for the next scan.
 */
static CTC_SERVICE_FN(ctc_conf_timing)
{
  CCBTimingCnf timing;  /* The new configuration */
  int arg=0;            /* The index of the argument being processed */
/*
 * Get the configuration arguments.
 */
  if(ctc_get_unsigned_scalar(ctc, "timing phase_switch_dt", argv[arg++],
			     NULL, &timing.phase_switch_dt, NULL) ||
     ctc_get_unsigned_scalar(ctc, "timing diode_rise_dt", argv[arg++],
			     &timing.diode_rise_dt, NULL, NULL) ||
     ctc_get_unsigned_scalar(ctc, "timing diode_fall_dt", argv[arg++],
			     &timing.diode_fall_dt, NULL, NULL) ||
     ctc_get_unsigned_scalar(ctc, "timing integ_period", argv[arg++],
			     &timing.integ_period, NULL, NULL) ||
     ctc_get_unsigned_scalar(ctc, "timing roundtrip_dt", argv[arg++],
			     NULL, &timing.roundtrip_dt, NULL) ||
     ctc_get_unsigned_scalar(ctc, "timing holdoff_dt", argv[arg++],
			     NULL, &timing.holdoff_dt, NULL))
    return TCL_ERROR;
/*
 * Install the new configuration.
 */
  if(ccb_set_timing_cnf(ctc->cnf, timing.phase_switch_dt, timing.diode_rise_dt,
			timing.diode_fall_dt, timing.integ_period,
			timing.roundtrip_dt, timing.holdoff_dt)) {
    Tcl_AppendResult(ctc->interp, "ccb timing: ", strerror(errno), ".", NULL);
    return TCL_ERROR;
  };
  return TCL_OK;
}

/*.......................................................................
 * Install a new sampler configuration for the next scan.
 */
static CTC_SERVICE_FN(ctc_conf_sampler)
{
  CCBSampleType type;    /* The type of samples to collect */
  const char *type_name; /* The textual name of the sample type */
/*
 * Get the configuration arguments.
 */
  type_name = argv[0];
  if(strcmp(type_name, "adc_samples") == 0) {
    type = CCB_ADC_SAMPLES;
  } else if(strcmp(type_name, "fake_samples") == 0) {
    type = CCB_FAKE_SAMPLES;
  } else {
    Tcl_AppendResult(ctc->interp,
		     "ccb sampler: Unknown sample_type argument: \"",
		     type_name, "\"", NULL);
    return TCL_ERROR;
  };
/*
 * Install the new configuration.
 */
  if(ccb_set_sampler_cnf(ctc->cnf, type)) {
    Tcl_AppendResult(ctc->interp, "ccb sampler: ", strerror(errno), ".", NULL);
    return TCL_ERROR;
  };
  return TCL_OK;
}

/*.......................................................................
 * Split a Tcl argument string into sub-arguments, attempt to read
 * each of these sub-arguments as unsigned integers, and assign them
 * to the specified array.
 *
 * Input:
 *  ctc       CCBTclClient *  The resource object of the CCB Tcl interface.
 *  name        const char *  The name of the target item, for use
 *                            in error messages.
 *  string      const char *  The Tcl argument string to be split.
 *  nmin               int    The minimum number of integers to read.
 *  nmax               int    The maximum number of integers to read.
 *  step               int    Allowable numbers of integers in the range nmin
 *                            to nmax are (nmin + step * n), where n is a
 *                            positive integer.
 * Input/Output:
 *  ulongs   unsigned long *  If an array is passed in this argument,
 *                            the values read will be recorded in it as
 *                            unsigned long elements. If not required,
 *                            pass NULL.
 *  ushorts unsigned short *  If an array is passed in this argument,
 *                            the values read will be recorded in it as
 *                            unsigned short elements. If not required,
 *                            pass NULL.
 *  uchars   unsigned char *  If an array is passed in this argument,
 *                            the values read will be recorded in it as
 *                            unsigned char elements. If not required,
 *                            pass NULL.
 *  nval          unsigned *  The number of values read.
 * Output:
 *  return             int    TCL_OK    - Success.
 *                            TCL_ERROR - Failure.
 */
static int ctc_get_unsigned_array(CCBTclClient *ctc, const char *name,
				  const char *string, int nmin, int nmax,
				  int step, unsigned long *ulongs,
				  unsigned short *ushorts,
				  unsigned char *uchars, unsigned *nval)
{
  int argc;    /* The number of sub-arguments placed in argv[] */
  char **argv; /* The array of argument strings extracted from string[] */
  int n = 0;   /* The number of integers successfully read */
  int status = TCL_OK; /* The return value of this function */
/*
 * Split the string into sub-argument strings.
 */
  if(Tcl_SplitList(ctc->interp, (char *)string, &argc, &argv) == TCL_ERROR)
    return TCL_ERROR;
/*
 * Check that the number of sub-arguments matches the requirements of
 * the caller.
 */
  if(argc < nmin || argc > nmax || (argc - nmin) % step != 0) {
    Tcl_AppendResult(ctc->interp, "Unacceptable number of elements for ",
		     (char *) name, ".", NULL);
    status = TCL_ERROR;
  };
/*
 * and record it in the variable associated with the pointer.
 */
  n = 0;
  while(n<argc && status==TCL_OK) {
    char *endp;
/*
 * Attempt to convert the latest argument to a number.
 */
    unsigned long number = strtoul(argv[n], &endp, 0);
/*
 * Check that the argument was converted successfully.
 */
    if(**argv == '\0' || *endp != '\0') {
      Tcl_AppendResult(ctc->interp, "Can't read ", (char *) name, ".", NULL);
      status = TCL_ERROR;
    } else {
/*
 * Record the new number.
 */
      if(ulongs)
	ulongs[n] = number;
      if(ushorts) {
	ushorts[n] = number;
	if((unsigned long) ushorts[n] != number) {
	  Tcl_AppendResult(ctc->interp, "An element of: ", (char *)name,
			   " is too big\n", NULL);
	  return TCL_ERROR;
	};
      };
      if(uchars) {
	uchars[n] = number;
	if((unsigned long) uchars[n] != number) {
	  Tcl_AppendResult(ctc->interp, "An element of: ", (char *)name,
			   " is too big\n", NULL);
	  return TCL_ERROR;
	};
      };

/*
 * Record the fact that a new integer has been converted and recorded
 * successfully.
 */
      n++;
    };
  };
/*
 * Return the number of successfully read integers.
 */
  if(nval)
    *nval = n;
/*
 * Delete the array of sub-argument strings.
 */
  Tcl_Free((char *) argv);
  return status;
}

/*.......................................................................
 * Return the contents of a monitor-data message via the Tcl result string.
 *
 * Input:
 *  ctc       CCBTclClient *  The resource object of the CCB Tcl interface.
 * Output:
 *  return             int    TCL_OK - The message contents can be found
 *                                     in the Tcl result string.
 *                            TCL_ERROR - An error occurred.
 */
static int ctc_get_monitor(CCBTclClient *ctc)
{
  CtcMonitorMsg *mon;   /* The most recently received monitor message */
  int i;
/*
 * Encode the contents of the most recent monitor message in the Tcl
 * result string, as a space-separated list of values.
 */
  mon = &ctc->messages.monitor;
  sprintf(ctc->buffer, "%lu %lu %lu %lu %lu %u", mon->ts.mjd, mon->ts.sec,
	  mon->ts.ns, mon->scan, mon->number, mon->nvalues);
  Tcl_AppendResult(ctc->interp, ctc->buffer, NULL);
  for(i=0; i<mon->nvalues; i++) {
    sprintf(ctc->buffer, " %lu", mon->values[i]);
    Tcl_AppendResult(ctc->interp, ctc->buffer, NULL);
  };
  return TCL_OK;
}

/*.......................................................................
 * Return the contents of a integrated-data message via the Tcl result
 * string.
 *
 * Input:
 *  ctc       CCBTclClient *  The resource object of the CCB Tcl interface.
 * Output:
 *  return             int    TCL_OK - The message contents can be found
 *                                     in the Tcl result string.
 *                            TCL_ERROR - An error occurred.
 */
static int ctc_get_integ(CCBTclClient *ctc)
{
  CtcIntegMsg *integ;   /* The most recently received integration message */
  int i;
/*
 * Encode the contents of the most recent integrated-data message in the Tcl
 * result string, as a space-separated list of values.
 */
  integ = &ctc->messages.integ;
  sprintf(ctc->buffer, "%lu %lu %lu %lu %lu {", integ->ts.mjd, integ->ts.sec,
	  integ->ts.ns, integ->scan, integ->number);
  Tcl_AppendResult(ctc->interp, ctc->buffer, NULL);
/*
 * List the integration status-flags.
 */
  if(integ->flags & CCB_CAL_A_ON)
     Tcl_AppendResult(ctc->interp, "cal_a_on ", NULL);
  if(integ->flags & CCB_CAL_B_ON)
     Tcl_AppendResult(ctc->interp, "cal_b_on ", NULL);
  if(integ->flags & CCB_INTEG_OK)
     Tcl_AppendResult(ctc->interp, "integ_ok ", NULL);
  if(integ->flags & CCB_SLAVE0_OK)
     Tcl_AppendResult(ctc->interp, "slave0_ok ",NULL);
  if(integ->flags & CCB_SLAVE1_OK)
     Tcl_AppendResult(ctc->interp, "slave1_ok ",NULL);
  if(integ->flags & CCB_SLAVE2_OK)
     Tcl_AppendResult(ctc->interp, "slave2_ok ",NULL);
  if(integ->flags & CCB_SLAVE3_OK)
     Tcl_AppendResult(ctc->interp, "slave3_ok ",NULL);
/*
 * Write the number of integrated values, and the values themselves.
 */
  sprintf(ctc->buffer, "} %u", integ->nvalues);
  Tcl_AppendResult(ctc->interp, ctc->buffer, NULL);
  for(i=0; i<integ->nvalues; i++) {
    sprintf(ctc->buffer, " %lu", integ->values[i]);
    Tcl_AppendResult(ctc->interp, ctc->buffer, NULL);
  };
  return TCL_OK;
}

/*.......................................................................
 * Return the contents of a log message via the Tcl result string.
 *
 * Input:
 *  ctc       CCBTclClient *  The resource object of the CCB Tcl interface.
 * Output:
 *  return             int    TCL_OK - The message contents can be found
 *                                     in the Tcl result string.
 *                            TCL_ERROR - An error occurred.
 */
static int ctc_get_log(CCBTclClient *ctc)
{
  CtcLogMsg *log;     /* The most recently received log message */
  const char *level;  /* The symbolic form of the log level */
/*
 * Encode the contents of the most recent integrated-data message in the Tcl
 * result string, as a space-separated list of values.
 */
  log = &ctc->messages.log;
  sprintf(ctc->buffer, "%lu %lu %lu %lu", log->ts.mjd, log->ts.sec, log->ts.ns,
	  log->id);
  Tcl_AppendResult(ctc->interp, ctc->buffer, NULL);
/*
 * Convert the log level to a symbolic form.
 */
  switch(log->level) {
  case CCB_INFO:
    level = "info";
    break;
  case CCB_NOTICE:
    level = "notice";
    break;
  case CCB_WARNING:
    level = "warning";
    break;
  case CCB_ERROR:
    level = "error";
    break;
  case CCB_FAULT:
    level = "fault";
    break;
  case CCB_FATAL:
    level = "fatal";
    break;
  default:
    level = "(unknown)";
    break;
  };
/*
 * Append the symbolic form of the level.
 */
  Tcl_AppendElement(ctc->interp, (char *)level);
  Tcl_AppendElement(ctc->interp, log->text);
  return TCL_OK;
}

/*.......................................................................
 * Return the contents of a status message via the Tcl result string.
 *
 * Input:
 *  ctc       CCBTclClient *  The resource object of the CCB Tcl interface.
 * Output:
 *  return             int    TCL_OK - The message contents can be found
 *                                     in the Tcl result string.
 *                            TCL_ERROR - An error occurred.
 */
static int ctc_get_status(CCBTclClient *ctc)
{
  unsigned long status;   /* A copy of the status being reported */
  unsigned reported = 0;  /* The set of reported status bits */
  int i;
/*
 * List all status bits and their symbolic names.
 */
  struct {
    unsigned bit;         /* The corresponding to the status bit */
    const char *name;     /* The symbolic name of the status bit */
  } status_bits[] = {
    {CCB_LINK_DOWN,   "link_down"},  {CCB_BUFFER_FULL, "buffer_full"},
    {CCB_HARD_FAULT,  "hard_fault"}, {CCB_SOFT_FAULT,  "soft_fault"}
  };
/*
 * Get a local copy of the status set.
 */
  status = ctc->messages.status;
/*
 * Check for the presence of each of the defined status bits, and
 * include them in the result string.
 */
  for(i=0; i<sizeof(status_bits)/sizeof(status_bits[0]); i++) {
    if(status_bits[i].bit & status) {
      reported |= status_bits[i].bit;
      Tcl_AppendElement(ctc->interp, (char *) status_bits[i].name);
    };
  };
/*
 * If any of the bits in 'status' weren't reported by above loop, this
 * implies that our list of possible status bits is incomplete.
 */
  if((status & ~reported) != 0) {
    Tcl_ResetResult(ctc->interp);
    Tcl_AppendResult(ctc->interp,
		     "Unknown status bit found in received CCB status.", NULL);
    return TCL_ERROR;
  };
  return TCL_OK;
}

/*.......................................................................
 * Return the contents of a command-error message via the Tcl result
 * string.
 *
 * Input:
 *  ctc       CCBTclClient *  The resource object of the CCB Tcl interface.
 * Output:
 *  return             int    TCL_OK - The message contents can be found
 *                                     in the Tcl result string.
 *                            TCL_ERROR - An error occurred.
 */
static int ctc_get_error(CCBTclClient *ctc)
{
  const char *cmderror;  /* The symbolic form of the completion status */
/*
 * Convert the status to a symbolic form.
 */
  switch(ctc->messages.cmderror) {
  case CCB_CMD_ACCEPTED:
    cmderror = "accepted";
    break;
  case CCB_CMD_GARBLED:
    cmderror = "garbled";
    break;
  case CCB_CMD_IGNORED:
    cmderror = "ignored";
    break;
  case CCB_CMD_SYSERR:
    cmderror = "syserr";
    break;
  default:
    Tcl_AppendResult(ctc->interp, "Unknown command error received.", NULL);
    return TCL_ERROR;
    break;
  };
/*
 * Append the symbolic form of the error status.
 */
  Tcl_AppendElement(ctc->interp, (char *)cmderror);
  return TCL_OK;
}

/*.......................................................................
 * Return the current data and time.
 *
 * Input:
 *  ctc       CCBTclClient *  The resource object of the CCB Tcl interface.
 * Output:
 *  return             int    TCL_OK - The message contents can be found
 *                                     in the Tcl result string.
 *                            TCL_ERROR - An error occurred.
 */
static int ctc_get_time(CCBTclClient *ctc)
{
  CCBTimeStamp t;  /* The current date and time */
/*
 * Get the current date and time.
 */
  if(ccb_get_timestamp(&t)) {
    Tcl_AppendResult(ctc->interp, "Error reading clock.", NULL);
    return TCL_ERROR;
  };
/*
 * Format the date and time in our buffer.
 */
  sprintf(ctc->buffer, "%lu %lu", t.mjd, t.sec);
/*
 * Copy the buffer to the Tcl result string.
 */
  Tcl_AppendResult(ctc->interp, ctc->buffer, NULL);
  return TCL_OK;
}

/*.......................................................................
 * The following function is called by the TCL event loop whenever the
 * the control socket becomes readbable or writable.
 *
 * Input:
 *  context   ClientData    The CCBTclClient resource object cast to
 *                          ClientData.
 *  mask             int    TCL_READABLE and/or TCL_WRITABLE.
 */
static void ctc_control_socket_handler(ClientData context, int mask)
{
  CCBTclClient *ctc = (CCBTclClient *) context;
  unsigned io;   /* The set of I/O to be performed */
/*
 * Compose the set of I/O directives to pass to ccb_client_communicate().
 */
  io = 0;
  if(mask & TCL_READABLE)
    io |= CCB_CNTRL_READ;
  if(mask & TCL_WRITABLE)
    io |= CCB_CNTRL_WRITE;
/*
 * Attempt to respond to the specified I/O events.
 */
  (void) ccb_client_communicate(ctc->client, io);
/*
 * Flush other event types so that we don't hog the Tcl event loop.
 */
  while(Tcl_DoOneEvent(TCL_WINDOW_EVENTS | TCL_TIMER_EVENTS |
		       TCL_IDLE_EVENTS | TCL_DONT_WAIT))
    ;
  return;
}

/*.......................................................................
 * The following function is called by the TCL event loop whenever the
 * the telemetry socket becomes readbable or writable.
 *
 * Input:
 *  context   ClientData    The CCBTclClient resource object cast to
 *                          ClientData.
 *  mask             int    TCL_READABLE and/or TCL_WRITABLE.
 */
static void ctc_telemetry_socket_handler(ClientData context, int mask)
{
  CCBTclClient *ctc = (CCBTclClient *) context;
  unsigned io;   /* The set of I/O to be performed */
/*
 * Compose the set of I/O directives to pass to ccb_client_communicate().
 */
  io = 0;
  if(mask & TCL_READABLE)
    io |= CCB_TELEM_READ;
  if(mask & TCL_WRITABLE)
    io |= CCB_TELEM_WRITE;
/*
 * Attempt to respond to the specified I/O events.
 */
  (void) ccb_client_communicate(ctc->client, io);
/*
 * Flush other event types so that we don't hog the Tcl event loop.
 */
  while(Tcl_DoOneEvent(TCL_WINDOW_EVENTS | TCL_TIMER_EVENTS |
		       TCL_IDLE_EVENTS | TCL_DONT_WAIT))
    ;
  return;
}

/*.......................................................................
 * This function is called by the CCB communications library to inform
 * us both of changes in the telementry and control sockets, and of
 * changes of the types of I/O currently expected for these sockets.
 */
static CCB_CLIENT_SOCKETS_FN(ctc_client_sockets_fn)
{
  CCBTclClient *ctc = (CCBTclClient *) data;
/*
 * If the control socket has changed, delete any associated file handler
 */
  if(cntrl_sock != ctc->cntrl_sock && ctc->cntrl_sock >= 0) {
    Tcl_DeleteFileHandler(ctc->cntrl_sock);
    ctc->cntrl_sock = -1;
  };
/*
 * If the telemetry socket has changed, delete any associated file handler
 */
  if(telem_sock != ctc->telem_sock && ctc->telem_sock >= 0) {
    Tcl_DeleteFileHandler(ctc->telem_sock);
    ctc->telem_sock = -1;
  };
/*
 * Update the control-socket file handler.
 */
  if(cntrl_sock >= 0) {
    ctc->cntrl_sock = cntrl_sock;
    Tcl_CreateFileHandler(cntrl_sock, 
			  ((io & CCB_CNTRL_READ) ? TCL_READABLE : 0) |
			  ((io & CCB_CNTRL_WRITE) ? TCL_WRITABLE : 0),
			  ctc_control_socket_handler, (ClientData) ctc);
  };
/*
 * Update the telemetry-socket file handler.
 */
  if(cntrl_sock >= 0) {
    ctc->telem_sock = telem_sock;
    Tcl_CreateFileHandler(telem_sock, 
			  ((io & CCB_TELEM_READ) ? TCL_READABLE : 0) |
			  ((io & CCB_TELEM_WRITE) ? TCL_WRITABLE : 0),
			  ctc_telemetry_socket_handler, (ClientData) ctc);
  };
  return 0;
}

/*.......................................................................
 * This is the function that ccb_client_communicate() calls whenever a
 * remote CCB server reports that a previously sent CCB control
 * command had a bad completion status.
 */
static CCB_CMD_ERROR_FN(ctc_cmd_error_fn)
{
  CCBTclClient *ctc = (CCBTclClient *) data;
/*
 * Make a record of the contents of the message, so that they can be
 * retrieved via the "ccb get error" command.
 */
  ctc->messages.cmderror = status;
/*
 * Inform the Tcl application that a new command-error is awaiting pickup.
 */
  return ctc->scripts.error ?
    Tcl_Eval(ctc->interp, ctc->scripts.error)!=TCL_OK : 0;
}

/*.......................................................................
 * This is the function that ccb_client_communicate() calls whenever a
 * reply to a preceding status-request command is received from a remote
 * CCB server.
 */
static CCB_STATUS_REPLY_FN(ctc_status_reply_fn)
{
  CCBTclClient *ctc = (CCBTclClient *) data;
/*
 * Make a record of the contents of the message, so that they can be
 * retrieved via the "ccb get status" command.
 */
  ctc->messages.status = status;
/*
 * Inform the Tcl application that a new status-reply is awaiting pickup.
 */
  return ctc->scripts.status ?
    Tcl_Eval(ctc->interp, ctc->scripts.status)!=TCL_OK : 0;
}

/*.......................................................................
 * This is the function that ccb_client_communicate() calls whenever a
 * new monitor-data message is received from a remote CCB server.
 */
static CCB_MONITOR_MSG_FN(ctc_monitor_msg_fn)
{
  CCBTclClient *ctc = (CCBTclClient *) data;
  CtcMonitorMsg *monitor = &ctc->messages.monitor;
  int i;
/*
 * Make a record of the contents of the message, so that they can be
 * retrieved via the "ccb get monitor" command.
 */
  monitor->ts = *ts;
  monitor->scan = scan;
  monitor->number = number;
  for(i=0; i<nvalues; i++)
    monitor->values[i] = values[i];
  monitor->nvalues = nvalues;
/*
 * Inform the Tcl application that a new monitor-data message
 * is awaiting pickup.
 */
  return ctc->scripts.monitor ?
    Tcl_Eval(ctc->interp, ctc->scripts.monitor)!=TCL_OK : 0;
}

/*.......................................................................
 * This is the function that ccb_client_communicate() calls whenever a
 * new integrated data message is received from a remote CCB server.
 */
static CCB_INTEG_MSG_FN(ctc_integ_msg_fn)
{
  CCBTclClient *ctc = (CCBTclClient *) data;
  CtcIntegMsg *integ = &ctc->messages.integ;
  int i;
/*
 * Make a record of the contents of the message, so that they can be
 * retrieved via the "ccb get integ" command.
 */
  integ->ts = *ts;
  integ->scan = scan;
  integ->number = number;
  integ->flags = flags;
  for(i=0; i<nvalues; i++)
    integ->values[i] = values[i];
  integ->nvalues = nvalues;
/*
 * Inform the Tcl application that a new integrated-data message is
 * awaiting pickup.
 */
  return ctc->scripts.integ ?
    Tcl_Eval(ctc->interp, ctc->scripts.integ)!=TCL_OK : 0;
}

/*.......................................................................
 * This is the function that ccb_client_communicate() calls whenever a
 * log message is received from a remote CCB server.
 */
static CCB_LOG_MSG_FN(ctc_log_msg_fn)
{
  CCBTclClient *ctc = (CCBTclClient *) data;
  CtcLogMsg *log = &ctc->messages.log;
/*
 * Make a record of the contents of the message, so that they can be
 * retrieved via the "ccb get log" command.
 */
  strncpy(log->text, msg, CCB_MAX_LOG);
  log->text[CCB_MAX_LOG-1] = '\0';
  log->id = id;
  log->level = level;
/*
 * Inform the Tcl application that a new command-error is awaiting pickup.
 */
  return ctc->scripts.log ?
    Tcl_Eval(ctc->interp, ctc->scripts.log)!=TCL_OK : 0;
}

/*.......................................................................
 * Get the bitwise union of the unsigned short values associated with a
 * space separated list of enumerator names.
 *
 * Input:
 *  ctc           CCBTclClient *  The resource object of the CCB Tcl
 *                                interface.
 *  command         const char *  The name of the originating Tcl command,
 *                                for use in error messages.
 *  type_name       const char *  The name by which to refer to members of
 *                                the set in error messages.
 *  string          const char *  The string containing the list of
 *                                enuemrator names.
 *  sym  const CtcUShortSymbol *  The synbol table of associations
 *                                between names and values.
 *  nsym                   int    The number of entries in sym[].
 * Input/Output:
 *  mask        unsigned short *  The bitwise union of the values
 *                                corresponding to the names found in
 *                                string[] will be assigned to *mask.
 * Output:
 *  return                 int    TCL_OK - The message contents can be found
 *                                     in the Tcl result string.
 *                                TCL_ERROR - An error occurred.
 */
static int ctc_get_ushort_bitmask(CCBTclClient *ctc, const char *command,
				  const char *type_name, const char *string,
				  const CtcUShortSymbol *sym, int nsym,
				  unsigned short *mask)
{
  const char *sptr;  /* A pointer into string[] */
  int i;
/*
 * Start with an empty bit-mask.
 */
  *mask = 0;
/*
 * Iterate through the input string, looking for words that name set
 * members to be included within the returned mask.
 */
  sptr=string;
  while(*sptr) {
/*
 * Find the start of the next word.
 */
    while(*sptr && isspace((int)(unsigned char) *sptr))
      sptr++;
/*
 * Check that we didn't hit the end of the string.
 */
    if(*sptr != '\0') {
      const char *wptr = sptr;  /* The start of the new word */
      size_t wlen;              /* The length of the new word */
/*
 * Find the end of the word.
 */
      while(*sptr && !isspace((int)(unsigned char) *sptr))
	sptr++;
/*
 * Compute the length of the word.
 */
      wlen = sptr - wptr;
/*
 * Copy the word into our buffer.
 */
      if(wlen >= CTC_BUF_SIZE) {
	Tcl_AppendResult(ctc->interp, (char *)command, ": Unexpectedly long ",
			 (char *)type_name, " name.", NULL);
	return TCL_ERROR;
      };
      strncpy(ctc->buffer, wptr, wlen);
      ctc->buffer[wlen] = '\0';
/*
 * Compare the new word to each of the known member names.
 */
      for(i=0; i<nsym && strcmp(ctc->buffer, sym[i].name) != 0; i++)
	;
      if(i < nsym) {
	*mask |= sym[i].value;
      } else {
	Tcl_AppendResult(ctc->interp, (char *)command, ": Unknown ",
			 (char *)type_name, " name: ", ctc->buffer, NULL);
	return TCL_ERROR;
      };
    };
  };
  return TCL_OK;
}

/*.......................................................................
 * Parse a space-separated list of stream names, returning the corresponding
 * bitset union of CCBTelemetryStream enumerators.
 *
 * Input:
 *  ctc        CCBTclClient *  The resource object of the CCB Tcl
 *                             interface.
 *  string       const char *  The string containing the list of
 *                             streams.
 * Input/Output:
 *  streams  unsigned short *  The bitwise union of CCBTelemetryStream
 *                             enumerators will be assigned to *streams.
 * Output:
 *  return              int    TCL_OK - The message contents can be found
 *                                      in the Tcl result string.
 *                             TCL_ERROR - An error occurred.
 */
static int ctc_get_stream_mask(CCBTclClient *ctc, const char *command,
			       const char *string, unsigned short *streams)
{
  return ctc_get_ushort_bitmask(ctc, command, "stream", string,
				ctc_stream_symbol_table,
				ctc_stream_symbol_count, streams);
}

/*.......................................................................
 * Parse a space-separated list of cal-diode names, returning the
 * corresponding bitset union of CCBCalDiodes enumerators.
 *
 * Input:
 *  ctc        CCBTclClient *  The resource object of the CCB Tcl
 *                             interface.
 *  string       const char *  The string containing the list of
 *                             cal-diode names.
 * Input/Output:
 *  diodes   unsigned short *  The bitwise union of CCBCalDiodes
 *                             enumerators will be assigned to *diodes.
 * Output:
 *  return              int    TCL_OK - The message contents can be found
 *                                      in the Tcl result string.
 *                             TCL_ERROR - An error occurred.
 */
static int ctc_get_cal_diode_mask(CCBTclClient *ctc, const char *command,
				  const char *string, unsigned short *diodes)
{
  return ctc_get_ushort_bitmask(ctc, command, "diode", string,
				ctc_diode_symbol_table,
				ctc_diode_symbol_count, diodes);
}

/*.......................................................................
 * Parse a space-separated list of phase-switch names, returning the
 * corresponding bitset union of CCBPhaseSwitches enumerators.
 *
 * Input:
 *  ctc         CCBTclClient *  The resource object of the CCB Tcl
 *                              interface.
 *  string        const char *  The string containing the list of
 *                              phase-switch names.
 * Input/Output:
 *  switches  unsigned short *  The bitwise union of CCBPhaseSwitches
 *                              enumerators will be assigned to *switches.
 * Output:
 *  return               int    TCL_OK - The message contents can be found
 *                                       in the Tcl result string.
 *                              TCL_ERROR - An error occurred.
 */
static int ctc_get_phase_switch_mask(CCBTclClient *ctc, const char *command,
				     const char *string,
				     unsigned short *switches)
{
  return ctc_get_ushort_bitmask(ctc, command, "switch", string,
				ctc_switch_symbol_table,
				ctc_switch_symbol_count, switches);
}
